# -*- coding: utf-8 -*-
from __future__ import print_function
import argparse
import logging
import importlib
import inspect
import os
import re
import six
import traceback
import yaml

from collections import defaultdict
from contextlib import contextmanager

import IPython
from pygments.console import ansiformat
from traitlets.config.loader import Config

from insights.core.context import SerializedArchiveContext
from insights.core.serde import Hydration
from insights.parsr.query import *  # noqa
from insights.parsr.query import eq, matches, make_child_query as q  # noqa
from insights.parsr.query.boolean import FALSE, TRUE

from insights import (
    apply_configs,
    apply_default_enabled,
    create_context,
    datasource,
    dr,
    extract,
    load_default_plugins,
    load_packages,
    parse_plugins,
)
from insights.core import plugins
from insights.core.context import HostContext
from insights.core.spec_factory import ContentProvider, RegistryPoint
from insights.formats import render
from insights.formats.text import render_links

Loader = getattr(yaml, "CSafeLoader", yaml.SafeLoader)

RULE_COLORS = {"fail": "brightred", "pass": "blue", "info": "magenta", "skip": "yellow"}


@contextmanager
def _create_new_broker(path=None):
    """
    Create a broker and populate it by evaluating the path with all
    registered datasources.

    Args:
        path (str): path to the archive or directory to analyze. ``None`` will
            analyze the current system.
    """
    datasources = dr.get_components_of_type(datasource)

    def make_broker(ctx):
        broker = dr.Broker()
        broker[ctx.__class__] = ctx

        if isinstance(ctx, SerializedArchiveContext):
            h = Hydration(ctx.root)
            broker = h.hydrate(broker=broker)

        dr.run(datasources, broker=broker)

        del broker[ctx.__class__]
        return broker

    if path:
        if os.path.isdir(path):
            ctx = create_context(path)
            yield (path, make_broker(ctx))
        else:
            with extract(path) as e:
                ctx = create_context(e.tmp_dir)
                yield (e.tmp_dir, make_broker(ctx))
    else:
        yield (os.curdir, make_broker(HostContext()))


# contextlib.ExitStack isn't available in all python versions
# so recurse to victory.
def with_brokers(archives, callback):
    brokers = []

    def inner(paths):
        if paths:
            path = paths.pop()
            with _create_new_broker(path) as ctx:
                brokers.append(ctx)
                inner(paths)
        else:
            callback(brokers)

    if archives:
        inner(list(reversed(archives)))
    else:
        with _create_new_broker() as ctx:
            callback([ctx])


def _get_available_models(broker, group=dr.GROUPS.single):
    """
    Given a broker populated with datasources, return everything that could
    run based on them.
    """
    state = set(broker.instances.keys())
    models = {}

    for comp in dr.run_order(dr.COMPONENTS[group]):
        if comp in dr.DELEGATES and not plugins.is_datasource(comp):
            if dr.DELEGATES[comp].get_missing_dependencies(state):
                continue

            if plugins.is_type(
                comp, (plugins.rule, plugins.condition, plugins.incident)
            ):
                name = "_".join(
                    [dr.get_base_module_name(comp), dr.get_simple_name(comp)]
                )
            else:
                name = dr.get_simple_name(comp)

            if name in models:
                prev = models[name]
                models[dr.get_name(prev).replace(".", "_")] = prev

                del models[name]
                name = dr.get_name(comp).replace(".", "_")

            models[name] = comp
            state.add(comp)

    return models


class Models(dict):
    u"""
    Represents all components that may be available given the data being
    analyzed. Use models.find() to see them. Tab complete attributes to access
    them. Use help(models) for more info.

    Start the shell with the environment variable
    INSIGHTS_FILTERS_ENABLED=False to disable filtering that may cause
    unexpected missing data.

    Examples:

        >>> models.find("rpm")
        InstalledRpms (insights.parsers.installed_rpms.InstalledRpms)

        >>> rpms = models.InstalledRpms
        >>> rpms.newest("bash")
        0:bash-4.1.2-48.el6

        >>> models.find("(?i)yum")  # Prefix "(?i)" ignores case.
        YumConf (insights.parsers.yum_conf.YumConf, parser)
        YumLog (insights.parsers.yumlog.YumLog, parser)
        YumRepoList (insights.parsers.yum.YumRepoList, parser)
        └╌╌Incorrect line: 'repolist: 33296'
        YumReposD (insights.parsers.yum_repos_d.YumReposD, parser)

        >>> models.show_trees("rpm")
        insights.parsers.installed_rpms.InstalledRpms (parser)
        ┊   insights.specs.Specs.installed_rpms (unfiltered / lines)
        ┊   ┊╌╌╌╌╌TextFileProvider("'/home/csams/Downloads/archives/sosreport-example-20191225000000/sos_commands/rpm/package-data'")
        ┊   ┊   insights.specs.insights_archive.InsightsArchiveSpecs.installed_rpms (unfiltered / lines)
        ┊   ┊   ┊   insights.specs.insights_archive.InsightsArchiveSpecs.all_installed_rpms (unfiltered / lines)
        ┊   ┊   ┊   ┊   insights.core.context.HostArchiveContext ()
        ┊   ┊   insights.specs.sos_archive.SosSpecs.installed_rpms (unfiltered / lines)
        ┊   ┊   ┊╌╌╌╌╌TextFileProvider("'/home/csams/Downloads/archives/sosreport-example-20191225000000/sos_commands/rpm/package-data'")
        ┊   ┊   ┊   insights.core.context.SosArchiveContext ()
        ┊   ┊   insights.specs.default.DefaultSpecs.installed_rpms (unfiltered / lines)
        ┊   ┊   ┊   insights.specs.default.DefaultSpecs.docker_installed_rpms (unfiltered / lines)
        ┊   ┊   ┊   ┊   insights.core.context.DockerImageContext ()
        ┊   ┊   ┊   insights.specs.default.DefaultSpecs.host_installed_rpms (unfiltered / lines)
        ┊   ┊   ┊   ┊   insights.core.context.HostContext ()
    """

    def __init__(self, broker, models, cwd, tmp, cov):
        self._requested = set()
        self._broker = broker
        self._cwd = cwd
        self._tmp = tmp
        self._cov = cov
        super(Models, self).__init__(models)

    def __dir__(self):
        """ Enabled ipython autocomplete. """
        return sorted(set(list(self.keys()) + dir(Models)))

    def _ipython_key_completions_(self):
        """ For autocomplete of keys when accessing models as a dict. """
        return sorted(self.keys())

    def __str__(self):
        return "{} components possibly available".format(len(self))

    def _get_color(self, comp):
        if comp in self._broker:
            if plugins.is_type(comp, plugins.rule) and self._broker[comp].get("type") == "skip":
                return "yellow"
            return "green"
        elif comp in self._broker.exceptions:
            return "brightred"
        elif comp in self._broker.missing_requirements:
            return "yellow"
        else:
            return ""

    def _dump_diagnostics(self, comp):
        results = []
        results.append("Dependency Tree")
        results.append("===============")
        results.extend(self._show_tree(comp))
        results.append("")
        results.append("Missing Dependencies")
        results.append("====================")
        results.extend(self._show_missing(comp))
        results.append("")
        results.append("Exceptions")
        results.append("==========")
        results.extend(self._show_exceptions(comp))
        IPython.core.page.page(six.u(os.linesep.join(results)))

    def evaluate_all(self, match=None, ignore="spec"):
        """
        Evaluate all components that match.

        Args:
            match (str, optional): regular expression for matching against
                the fully qualified name of components to keep.
            ignore (str, optional): regular expression for searching against
                the fully qualified name of components to ignore.
        """
        match, ignore = self._desugar_match_ignore(match, ignore)

        tasks = []
        for c in self.values():
            name = dr.get_name(c)
            if match.test(name) and not ignore.test(name):
                if not any(
                    [
                        c in self._broker.instances,
                        c in self._broker.exceptions,
                        c in self._broker.missing_requirements,
                    ]
                ):
                    tasks.append(c)

        if not tasks:
            return

        dr.run(tasks, broker=self._broker)
        self.show_timings(match, ignore)

    def evaluate(self, name):
        """
        Evaluate a component and return its result. Prints diagnostic
        information in the case of failure. This function is useful when a
        component's name contains characters that aren't valid for python
        identifiers so you can't access it with ``models.<name>``.

        Args:
            name (str): the name of the component as shown by :func:`Models.find()`.
        """
        comp = self.get(name) or dr.get_component(name)
        if not comp:
            return

        if not plugins.is_rule(comp):
            self._requested.add((name, comp))

        if comp in self._broker:
            return self._broker[comp]

        if comp in self._broker.exceptions or comp in self._broker.missing_requirements:
            self._dump_diagnostics(comp)
            return

        val = dr.run(comp, broker=self._broker).get(comp)

        if comp not in self._broker:
            if comp in self._broker.exceptions or comp in self._broker.missing_requirements:
                self._dump_diagnostics(comp)
            else:
                print("{} chose to skip.".format(dr.get_name(comp)))
        return val

    def __getattr__(self, name):
        return self.evaluate(name)

    # TODO: lot of room for improvement here...
    def make_rule(self, path=None, overwrite=False, pick=None):
        """
        Attempt to generate a rule based on models used so far.

        Args:
            path(str): path to store the rule.
            overwrite (bool): whether to overwrite an existing file.
            pick (str): Optionally specify which lines or line ranges
                to use for the rule body. "1 2 3" gets lines 1,2 and 3.
                "1 3-5 7" gets line 1, lines 3 through 5, and line 7.
        """
        import IPython

        ip = IPython.get_ipython()
        ignore = [
            r"=.*models\.",
            r"^(%|!|help)",
            r"make_rule",
            r"models\.(show|find).*",
            r".*\?$",
            r"^(clear|pwd|cd *.*|ll|ls)$",
        ]

        if pick:
            lines = [r[2] for r in ip.history_manager.get_range_by_str(pick)]
        else:
            lines = []
            for r in ip.history_manager.get_range():
                l = r[2]
                if any(re.search(i, l) for i in ignore):
                    continue
                elif l.startswith("models."):
                    l = l[7:]
                lines.append(l)

        # the user asked for these models during the session.
        requested = sorted(self._requested, key=lambda i: i[0])

        # figure out what we need to import for filtering.
        filterable = defaultdict(list)
        for _, c in requested:
            for d in dr.get_dependency_graph(c):
                try:
                    if isinstance(d, RegistryPoint) and d.filterable:
                        n = d.__qualname__
                        cls = n.split(".")[0]
                        filterable[dr.get_module_name(d)].append((cls, n))
                except:
                    pass

        model_names = [dr.get_simple_name(i[1]) for i in requested]
        var_names = [i[0].lower() for i in requested]

        imports = [
            "from insights import rule, make_fail, make_info, make_pass  # noqa",
            "from insights.parsr.query import *  # noqa",
            "from insights.parsr.query import make_child_query as q  # noqa",
            "",
        ]

        for _, c in requested:
            mod = dr.get_module_name(c)
            name = dr.get_simple_name(c)
            imports.append("from {} import {}".format(mod, name))

        seen = set()
        if filterable:
            imports.append("")

        for k in sorted(filterable):
            for (cls, n) in filterable[k]:
                if (k, cls) not in seen:
                    imports.append("from {} import {}".format(k, cls))
                    seen.add((k, cls))

        seen = set()
        filters = []
        for k in sorted(filterable):
            for i in filterable[k]:
                if i not in seen:
                    filters.append("add_filter({}, ...)".format(i[1]))
                    seen.add(i)

        if filters:
            imports.append("from insights.core.filters import add_filter")

        filter_stanza = os.linesep.join(filters)
        import_stanza = os.linesep.join(imports)
        decorator = "@rule({})".format(", ".join(model_names))
        func_decl = "def report({}):".format(", ".join(var_names))
        body = os.linesep.join(["    " + x for x in lines]) if lines else "    pass"

        res = import_stanza
        if filter_stanza:
            res += (os.linesep * 2) + filter_stanza

        res += (os.linesep * 3) + decorator + os.linesep + func_decl + os.linesep + body

        if path:
            if not path.startswith("/"):
                realpath = os.path.realpath(path)
                if not realpath.startswith(self._cwd):
                    path = os.path.join(self._cwd, path)

            if os.path.exists(path) and not overwrite:
                print(
                    "{} already exists. Use overwrite=True to overwrite.".format(path)
                )
                return

            if not os.path.exists(path) or overwrite:
                with open(path, "w") as f:
                    f.write(res)
                ip.magic("edit -x {}".format(path))
                print("Saved to {}".format(path))
        else:
            IPython.core.page.page(ip.pycolorize(res))

    def _desugar_match_ignore(self, match, ignore):
        if match is None:
            match = TRUE
        elif isinstance(match, str):
            if match in self:
                match = eq(dr.get_name(self[match]))
            else:
                match = matches(match)

        if ignore is None:
            ignore = FALSE
        elif isinstance(ignore, str):
            ignore = matches(ignore)

        return (match, ignore)

    def _show_missing(self, comp):
        results = []
        try:
            req, alo = self._broker.missing_requirements[comp]
            name = dr.get_name(comp)
            results.append(name)
            results.append("-" * len(name))
            if req:
                results.append("Requires:")
                for r in req:
                    results.append("    {}".format(dr.get_name(r)))
            if alo:
                if req:
                    results.append("")
                results.append("Requires At Least One From Each List:")
                for r in alo:
                    results.append("[")
                    for i in r:
                        results.append("    {}".format(dr.get_name(i)))
                    results.append("]")
        except:
            pass
        return results

    def show_requested(self):
        """ Show the components you've worked with so far. """
        results = []
        for name, comp in sorted(self._requested):
            results.append(
                ansiformat(
                    self._get_color(comp), "{} {}".format(name, dr.get_name(comp))
                )
            )
        IPython.core.page.page(six.u(os.linesep.join(results)))

    def reset_requested(self):
        """ Reset requested state so you can work on a new rule. """
        IPython.get_ipython().history_manager.reset()
        self._requested.clear()

    def show_source(self, comp):
        """
        Show source for the given module, class, or function. Also accepts
        the string names, with the side effect that the component will be
        imported.
        """
        try:
            if isinstance(comp, six.string_types):
                comp = self.get(comp) or dr.get_component(comp) or importlib.import_module(comp)
            comp = inspect.getmodule(comp)
            ip = IPython.get_ipython()
            if self._cov:
                path, runnable, excluded, not_run, _ = self._cov.analysis2(comp)
                runnable, not_run = set(runnable), set(not_run)
                src = ip.pycolorize(inspect.getsource(comp)).splitlines()
                width = len(str(len(src)))
                template = "{0:>%s}" % width
                results = []
                file_line = "{} {}".format(
                    ansiformat("red", "File:"), os.path.realpath(path)
                )
                explain_line = "{} numbered lines have executed. python standard libs are excluded.".format(
                    ansiformat("*brightgreen*", "Green")
                )
                results.append(file_line)
                results.append(explain_line)
                results.append("")
                for i, line in enumerate(src, start=1):
                    prefix = template.format(i)
                    if i in runnable and i not in not_run:
                        color = "*brightgreen*"
                    else:
                        color = "gray"
                    results.append("{} {}".format(ansiformat(color, prefix), line))
                IPython.core.page.page(six.u(os.linesep.join(results)))
            else:
                ip.inspector.pinfo(comp, detail_level=1)
        except:
            traceback.print_exc()

    def _get_type_name(self, comp):
        try:
            return dr.get_component_type(comp).__name__
        except:
            return ""

    def _get_rule_value_kind(self, val):
        if val is None:
            kind = None
        elif isinstance(val, plugins.make_response):
            kind = "fail"
        elif isinstance(val, plugins.make_pass):
            kind = "pass"
        elif isinstance(val, plugins.make_info):
            kind = "info"
        elif isinstance(val, plugins._make_skip):
            kind = "skip"
        elif isinstance(val, plugins.Response):
            kind = val.response_type or ""
        else:
            kind = ""
        return kind

    def _get_rule_value(self, comp):
        try:
            val = self._broker[comp]
            if plugins.is_rule(comp):
                _type = val.__class__.__name__
                kind = self._get_rule_value_kind(val)
                color = RULE_COLORS.get(kind, "")
                return ansiformat(color, " [{}]".format(_type))
        except:
            pass
        return ""

    def _show_datasource(self, d, v, indent=""):
        try:
            filtered = "filtered" if dr.DELEGATES[d].filterable else "unfiltered"
            mode = "bytes" if dr.DELEGATES[d].raw else "lines"
        except:
            filtered = "unknown"
            mode = "unknown"

        desc = "{n} ({f} / {m})".format(n=dr.get_name(d), f=filtered, m=mode)
        color = self._get_color(d)

        results = []
        results.append(indent + ansiformat(color, desc))

        if not v:
            return results
        if not isinstance(v, list):
            v = [v]

        for i in v:
            if isinstance(i, ContentProvider):
                s = ansiformat(color, str(i))
            else:
                s = ansiformat(color, "<intermediate value>")
            results.append("{}\u250A\u254C\u254C\u254C\u254C\u254C{}".format(indent, s))
        return results

    def _show_tree(self, node, indent="", depth=None, dep_getter=dr.get_dependencies):
        if depth is not None and depth == 0:
            return []

        results = []
        color = self._get_color(node)
        if plugins.is_datasource(node):
            results.extend(
                self._show_datasource(node, self._broker.get(node), indent=indent)
            )
        else:
            _type = self._get_type_name(node)
            name = dr.get_name(node)
            suffix = self._get_rule_value(node)
            desc = ansiformat(color, "{n} ({t}".format(n=name, t=_type))
            results.append(indent + desc + suffix + ansiformat(color, ")"))

        dashes = "\u250A\u254C\u254C\u254C\u254C\u254C"
        if node in self._broker.exceptions:
            for ex in self._broker.exceptions[node]:
                results.append(indent + dashes + ansiformat(color, str(ex)))

        deps = dep_getter(node)
        next_indent = indent + "\u250A   "
        for d in deps:
            results.extend(
                self._show_tree(
                    d, next_indent, depth=depth if depth is None else depth - 1, dep_getter=dep_getter
                )
            )
        return results

    def show_trees(self, match=None, ignore="spec", depth=None, toward_dependents=False):
        """
        Show dependency trees of any components whether they're available or not.

        Args:
            match (str, optional): regular expression for matching against
                the fully qualified name of components to keep.
            ignore (str, optional): regular expression for searching against
                the fully qualified name of components to ignore.
            depth (int, optional): how deep into the tree to explore.
            toward_dependents (bool, optional): whether to walk the tree toward dependents.
                Default is to walk toward dependencies.
        """
        match, ignore = self._desugar_match_ignore(match, ignore)
        dep_getter = dr.get_dependents if toward_dependents else dr.get_dependencies

        graph = defaultdict(list)
        for c in dr.DELEGATES:
            name = dr.get_name(c)
            if match.test(name) and not ignore.test(name):
                graph[name].append(c)

        results = []
        for name in sorted(graph):
            for c in graph[name]:
                results.extend(self._show_tree(c, depth=depth, dep_getter=dep_getter))
                results.append("")
        IPython.core.page.page(six.u(os.linesep.join(results)))

    def show_failed(self, match=None, ignore="spec"):
        """
        Show names of any components that failed during evaluation. Ignores
        "spec" by default.

        Args:
            match (str, optional): regular expression for matching against
                the fully qualified name of components to keep.
            ignore (str, optional): regular expression for searching against
                the fully qualified name of components to ignore.
        """
        match, ignore = self._desugar_match_ignore(match, ignore)

        mid_dashes = "\u250A\u254C\u254C"
        bottom_dashes = "\u2514\u254C\u254C"
        results = []
        for comp in sorted(self._broker.exceptions, key=dr.get_name):
            name = dr.get_name(comp)
            if match.test(name) and not ignore.test(name):
                color = self._get_color(comp)
                results.append(ansiformat(color, name))
                exes = self._broker.exceptions[comp]
                last = len(exes) - 1
                for i, ex in enumerate(exes):
                    dashes = bottom_dashes if i == last else mid_dashes
                    results.append(ansiformat(color, dashes + str(ex)))
                results.append("")
        IPython.core.page.page(six.u(os.linesep.join(results)))

    def _show_exceptions(self, comp):
        name = dr.get_name(comp)
        results = [ansiformat("*brightred*", name)]
        results.append(ansiformat("*brightred*", "-" * len(name)))
        for e in self._broker.exceptions.get(comp, []):
            t = self._broker.tracebacks.get(e)
            if t:
                results.append(t)
        return results

    def show_exceptions(self, match=None, ignore="spec"):
        """
        Show exceptions that occurred during evaluation. Ignores "spec" by
        default.

        Args:
            match (str, optional): regular expression for matching against
                the fully qualified name of components to keep.
            ignore (str, optional): regular expression for searching against
                the fully qualified name of components to ignore.
        """
        match, ignore = self._desugar_match_ignore(match, ignore)

        results = []
        for comp in sorted(self._broker.exceptions, key=dr.get_name):
            name = dr.get_name(comp)
            if match.test(name) and not ignore.test(name):
                results.extend(self._show_exceptions(comp))
        IPython.core.page.page(six.u(os.linesep.join(results)))

    def show_rule_report(self, match=None, ignore=None):
        """
        Print a rule report for the matching rules.
        """
        match, ignore = self._desugar_match_ignore(match, ignore)
        results = defaultdict(dict)

        for comp, val in self._broker.items():
            name = dr.get_name(comp)
            if plugins.is_rule(comp) and match.test(name) and not ignore.test(name):
                kind = self._get_rule_value_kind(val)

                if kind:
                    body = render(comp, val)
                    links = render_links(comp)
                    results[kind][name] = os.linesep.join([body, "", links])

        report = []
        for kind in ["info", "pass", "fail"]:
            color = RULE_COLORS.get(kind, "")
            hits = results.get(kind, {})
            for name in sorted(hits):
                report.append(ansiformat(color, name))
                report.append(ansiformat(color, "-" * len(name)))
                report.append(hits[name])
                report.append("")
        IPython.core.page.page(six.u(os.linesep.join(report)))

    def show_timings(self, match=None, ignore="spec", group=dr.GROUPS.single):
        """
        Show timings for components that have successfully evaluated.

        Args:
            match (str, optional): regular expression for matching against
                the fully qualified name of components to keep.
            ignore (str, optional): regular expression for searching against
                the fully qualified name of components to ignore.
        """
        match, ignore = self._desugar_match_ignore(match, ignore)

        results = []
        total = 0.0
        for comp in dr.COMPONENTS[group]:
            name = dr.get_name(comp)
            if comp in self._broker.exec_times and match.test(name) and not ignore.test(name):
                color = self._get_color(comp)
                t = self._broker.exec_times[comp]
                total += t
                results.append((t, name, color))

        report = [ansiformat("brightmagenta", "Total: {:.10f} seconds".format(total)), ""]
        for timing, name, color in sorted(results, reverse=True):
            report.append(ansiformat(color, "{:.10f}: {}".format(timing, name)))

        IPython.core.page.page(six.u(os.linesep.join(report)))

    def find(self, match=None, ignore=None):
        """
        Find components that might be available based on the data being
        analyzed.

        Args:
            match (str, optional): regular expression for matching against
                the fully qualified name of components to keep.
            ignore (str, optional): regular expression for searching against
                the fully qualified name of components to ignore.
        """
        match, ignore = self._desugar_match_ignore(match, ignore)
        mid_dashes = "\u250A\u254C\u254C"
        bottom_dashes = "\u2514\u254C\u254C"
        results = []
        for p in sorted(self, key=str.lower):
            comp = self[p]
            name = dr.get_name(comp)
            if match.test(name) and not ignore.test(name):
                color = self._get_color(comp)
                _type = self._get_type_name(comp)
                suffix = self._get_rule_value(comp)
                desc = ansiformat(color, "{p} ({n}, {t}".format(p=p, n=name, t=_type))
                results.append(desc + suffix + ansiformat(color, ")"))
                if comp in self._broker.exceptions:
                    exes = self._broker.exceptions[comp]
                    last = len(exes) - 1
                    for i, ex in enumerate(exes):
                        dashes = bottom_dashes if i == last else mid_dashes
                        results.append(ansiformat(color, dashes + str(ex)))
        IPython.core.page.page(six.u(os.linesep.join(results)))


class Holder(dict):
    """
    This is a dictionary that holds models for multiple archives. Access each model
    set using the path to the archive as the key. See models.keys().
    """
    def _ipython_key_completions_(self):
        return self.keys()


def start_session(paths, change_directory=False, __coverage=None):
    __cwd = os.path.abspath(os.curdir)

    def callback(brokers):
        models = Holder()
        for i, (path, broker) in enumerate(brokers):
            avail = _get_available_models(broker)
            if paths:
                if len(paths) > 1:
                    models[paths[i]] = Models(broker, avail, __cwd, path, __coverage)
                else:
                    models = Models(broker, avail, __cwd, path, __coverage)
            else:
                models = Models(broker, avail, __cwd, path, __coverage)

        if change_directory and len(brokers) == 1:
            __working_path, _ = brokers[0]
            os.chdir(__working_path)
        # disable jedi since it won't autocomplete for objects with__getattr__
        # defined.
        IPython.core.completer.Completer.use_jedi = False
        __cfg = Config()
        __cfg.TerminalInteractiveShell.banner1 = Models.__doc__
        __ns = {}
        __ns.update(globals())
        __ns.update({"models": models})
        IPython.start_ipython([], user_ns=__ns, config=__cfg)

    with_brokers(paths, callback)
    if change_directory:
        os.chdir(__cwd)


def _handle_config(config):
    if config:
        with open(config) as f:
            cfg = yaml.load(f, Loader=Loader)
            load_packages(cfg.get("packages", []))
            apply_default_enabled(cfg)
            apply_configs(cfg)


def _parse_args():
    desc = "Perform interactive system analysis with insights components."
    epilog = """
        Set env INSIGHTS_FILTERS_ENABLED=False to disable filtering that may
        cause unexpected missing data.
    """.strip()
    p = argparse.ArgumentParser(description=desc, epilog=epilog)

    p.add_argument(
        "-p", "--plugins", default="", help="Comma separated list of packages to load."
    )
    p.add_argument("-c", "--config", help="The insights configuration to apply.")
    p.add_argument(
        "--no-coverage",
        action="store_true",
        help="Don't show code coverage when viewing source.",
    )
    p.add_argument(
        "--cd",
        action="store_true",
        help="Change into the expanded directory for analysis.",
    )
    p.add_argument(
        "--no-defaults", action="store_true", help="Don't load default components."
    )
    p.add_argument(
        "-v", "--verbose", action="store_true", help="Global debug level logging."
    )

    path_desc = "Archives or paths to analyze. Leave off to target the current system."
    p.add_argument("paths", nargs="*", help=path_desc)

    return p.parse_args()


def main():
    args = _parse_args()
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.ERROR)

    cov = None
    if not args.no_coverage:
        from coverage import Coverage

        cov = Coverage(cover_pylib=False)
        cov.start()

    if not args.no_defaults:
        load_default_plugins()
        dr.load_components("insights.parsers", "insights.combiners")

    load_packages(parse_plugins(args.plugins))
    _handle_config(args.config)

    start_session(args.paths, args.cd, __coverage=cov)
    if cov:
        cov.stop()
        cov.erase()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Sample Rule
===========

This is a simple rule and can be run against the local host
using the following command::

    $ insights-run -p examples.rules.sample_script

or from the examples/rules directory::

    $ ./sample_rules.py
"""
from insights.core.plugins import make_fail, make_pass, rule
from insights.parsers.redhat_release import RedhatRelease

# Jinga template for message to be displayed for either
# response tag
CONTENT = "This machine runs {{product}}."


@rule(RedhatRelease, content=CONTENT)
def report(rel):
    """Fires if the machine is running Fedora."""

    if "Fedora" in rel.product:
        return make_pass("IS_FEDORA", product=rel.product)
    else:
        return make_fail("IS_NOT_FEDORA", product=rel.product)


if __name__ == "__main__":
    from insights import run
    run(report, print_summary=True)



"""
Insights Core is a data collection and analysis framework that is built for
extensibility and rapid development. It includes a set of reusable components
for gathering data in myriad ways and providing a reliable object model for it.

.. code-block: python

    >>> from insights import run
    >>> from insights.parsers import installed_rpms as rpm
    >>> lower = rpm.Rpm("bash-4.4.11-1.fc26")
    >>> upper = rpm.Rpm("bash-4.4.22-1.fc26")
    >>> results = run(rpm.Installed)
    >>> rpms = results[rpm.Installed]
    >>> rpms.newest("bash")
    "0:bash-4.4.12-7.fc26"
    >>> lower <= rpms.newest("bash") < upper
    True
"""
from __future__ import print_function
import logging
import pkgutil
import os
import sys
import yaml
from collections import defaultdict

from .core import Scannable, LogFileOutput, Parser, IniConfigFile  # noqa: F401
from .core import FileListing, LegacyItemAccess, SysconfigOptions  # noqa: F401
from .core import YAMLParser, JSONParser, XMLParser, CommandParser  # noqa: F401
from .core import AttributeDict  # noqa: F401
from .core import Syslog  # noqa: F401
from .core import taglang
from .core.archives import COMPRESSION_TYPES, extract, InvalidArchive, InvalidContentType  # noqa: F401
from .core import dr  # noqa: F401
from .core.context import ClusterArchiveContext, HostContext, HostArchiveContext, SerializedArchiveContext, ExecutionContext  # noqa: F401
from .core.dr import SkipComponent  # noqa: F401
from .core.hydration import create_context, initialize_broker  # noqa: F401
from .core.plugins import combiner, fact, metadata, parser, rule  # noqa: F401
from .core.plugins import datasource, condition, incident  # noqa: F401
from .core.plugins import make_response, make_metadata, make_fingerprint  # noqa: F401
from .core.plugins import make_pass, make_fail, make_info  # noqa: F401
from .core.filters import add_filter, apply_filters, get_filters  # noqa: F401
from .formats import get_formatter
from .parsers import get_active_lines  # noqa: F401
from .util import defaults  # noqa: F401
from .formats import Formatter as FormatterClass

from .core.spec_factory import RawFileProvider, TextFileProvider

log = logging.getLogger(__name__)


package_info = dict((k, None) for k in ["RELEASE", "COMMIT", "VERSION", "NAME"])


for name in package_info:
    package_info[name] = pkgutil.get_data(__name__, name).strip().decode("utf-8")


def get_nvr():
    return "{0}-{1}-{2}".format(package_info["NAME"],
                                package_info["VERSION"],
                                package_info["RELEASE"])


RULES_STATUS = {}
"""
Mapping of dictionaries containing nvr and commitid for each rule repo included
in this instance

{"rule_repo_1": {"version": nvr(), "commit": sha1}}
"""


def add_status(name, nvr, commit=None):
    """
    Rule repositories should call this method in their package __init__ to
    register their version information.
    """
    RULES_STATUS[name] = {"version": nvr, "commit": commit}


add_status(package_info["NAME"], get_nvr(), package_info["COMMIT"])


def process_dir(broker, root, graph, context, inventory=None):
    ctx, broker = initialize_broker(root, context=context, broker=broker)
    log.debug("Processing %s with %s" % (root, ctx))

    if isinstance(ctx, ClusterArchiveContext):
        from .core.cluster import process_cluster
        archives = [f for f in ctx.all_files if f.endswith(COMPRESSION_TYPES)]
        return process_cluster(graph, archives, broker=broker, inventory=inventory)

    graph = dict((k, v) for k, v in graph.items() if k in dr.COMPONENTS[dr.GROUPS.single])
    broker = dr.run(graph, broker=broker)
    return broker


def _run(broker, graph=None, root=None, context=None, inventory=None):
    """
    run is a general interface that is meant for stand alone scripts to use
    when executing insights components.

    Args:
        root (str): None will causes a host collection in which command and
            file specs are run. A directory or archive path will cause
            collection from the directory or archive, and only file type specs
            or those that depend on `insights.core.context.HostArchiveContext`
            will execute.
        component (function or class): The component to execute. Will only execute
            the component and its dependency graph. If None, all components with
            met dependencies will execute.

    Returns:
        broker: object containing the result of the evaluation.
    """

    if not root:
        context = context or HostContext
        broker[context] = context()
        graph = dict((k, v) for k, v in graph.items() if k in dr.COMPONENTS[dr.GROUPS.single])
        return dr.run(graph, broker=broker)

    if os.path.isdir(root):
        return process_dir(broker, root, graph, context, inventory=inventory)
    else:
        with extract(root) as ex:
            return process_dir(broker, ex.tmp_dir, graph, context, inventory=inventory)


def load_default_plugins():
    dr.load_components("insights.specs.default")
    dr.load_components("insights.specs.insights_archive")
    dr.load_components("insights.specs.core3_archive")
    dr.load_components("insights.specs.sos_archive")
    dr.load_components("insights.specs.jdr_archive")
    dr.load_components("insights.specs.must_gather_archive")


def load_packages(packages):
    plugins = []
    for name in packages:
        if name not in sys.modules:
            plugins.append(name)
            dr.load_components(name, continue_on_error=False)

    return plugins


def parse_plugins(p):
    plugins = []
    if p:
        for path in p.split(","):
            path = path.strip()
            if path.endswith(".py"):
                path, _ = os.path.splitext(path)
            path = path.rstrip("/").replace("/", ".")
            plugins.append(path)
    return plugins


def apply_default_enabled(config):
    """
    Configures dr and already loaded components with a default enabled
    value.
    """
    default_enabled = config.get("default_component_enabled", True)
    for k in dr.ENABLED:
        dr.ENABLED[k] = default_enabled

    enabled = defaultdict(lambda: default_enabled)
    enabled.update(dr.ENABLED)
    dr.ENABLED = enabled


def apply_configs(config):
    """
    Configures components.

    Args:
        config (dict): a dictionary with the following keys:
            default_component_enabled (bool, optional): default value for
                whether components are enable if not specifically declared in
                the config section. Defaults to True.

            configs (list): list of dictionaries with the following keys:
                name, enabled, metadata, and timeout. All keys are optional
                except name.

                name is the prefix or exact name of any loaded component. Any
                component starting with name will have the associated
                configuration applied.

                enabled is whether the matching components will execute even if
                their dependencies are met. Defaults to True.

                timeout sets the class level timeout attribute of any component
                so long as the attribute already exists.

                metadata is any dictionary that you want to attach to the
                component. The dictionary can be retrieved by the component at
                runtime.
    """
    default_enabled = config.get("default_component_enabled", True)
    delegate_keys = sorted(dr.DELEGATES, key=dr.get_name)
    for comp_cfg in config.get("configs", []):
        name = comp_cfg.get("name")
        for c in delegate_keys:
            delegate = dr.DELEGATES[c]
            cname = dr.get_name(c)
            if cname.startswith(name):
                dr.ENABLED[c] = comp_cfg.get("enabled", default_enabled)
                delegate.metadata.update(comp_cfg.get("metadata", {}))
                delegate.tags = set(comp_cfg.get("tags", delegate.tags))
                for k, v in delegate.metadata.items():
                    if hasattr(c, k):
                        log.debug("Setting %s.%s to %s", cname, k, v)
                        setattr(c, k, v)

                if hasattr(c, "timeout"):
                    c.timeout = comp_cfg.get("timeout", c.timeout)

                if hasattr(delegate, "links"):
                    delegate.links = comp_cfg.get("links", delegate.links)
            if cname == name:
                break


def _load_context(path):
    if path is None:
        return

    if "." not in path:
        path = ".".join(["insights.core.context", path])
    return dr.get_component(path)


def run(component=None, root=None, print_summary=False,
        context=None, inventory=None, print_component=None):

    load_default_plugins()

    args = None
    formatter = None
    formatters = None
    if print_summary:
        import argparse
        import logging
        p = argparse.ArgumentParser(add_help=False)
        p.add_argument("archive", nargs="?", help="Archive or directory to analyze.")
        p.add_argument("-p", "--plugins", default="",
                       help="Comma-separated list without spaces of package(s) or module(s) containing plugins.")
        p.add_argument("-b", "--bare",
                       help='Specify "spec=filename[,spec=filename,...]" to use the bare file for the spec',
                       default="")
        p.add_argument("-c", "--config", help="Configure components.")
        p.add_argument("-i", "--inventory", help="Ansible inventory file for cluster analysis.")
        p.add_argument("-k", "--pkg-query", help="Expression to select rules by package.")
        p.add_argument("-v", "--verbose", help="Verbose output.", action="store_true")
        p.add_argument("-f", "--format", help="Output format.", default="insights.formats.text")
        p.add_argument("-s", "--syslog", help="Log results to syslog.", action="store_true")
        p.add_argument("--tags", help="Expression to select rules by tag.")
        p.add_argument("-D", "--debug", help="Verbose debug output.", action="store_true")
        p.add_argument("--context", help="Execution Context. Defaults to HostContext if an archive isn't passed.")

        class Args(object):
            pass

        formatters = []
        args = Args()
        p.parse_known_args(namespace=args)
        p = argparse.ArgumentParser(parents=[p])
        args.format = "insights.formats._json" if args.format == "json" else args.format
        args.format = "insights.formats._yaml" if args.format == "yaml" else args.format
        fmt = args.format if "." in args.format else "insights.formats." + args.format
        Formatter = dr.get_component(fmt)
        if not Formatter or not isinstance(Formatter, FormatterClass):
            dr.load_components(fmt, continue_on_error=False)
            Formatter = get_formatter(fmt)
        Formatter.configure(p)
        p.parse_args(namespace=args)
        formatter = Formatter(args)
        formatters.append(formatter)

        if args.syslog:
            fmt = "insights.formats._syslog"
            Formatter = dr.get_component(fmt)
            if not Formatter:
                dr.load_components(fmt, continue_on_error=False)
                Formatter = get_formatter(fmt)
            p.parse_args(namespace=args)
            formatter = Formatter(args)
            formatters.append(formatter)

        logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO if args.verbose else logging.ERROR)
        context = _load_context(args.context) or context
        inventory = args.inventory

        root = args.archive or root
        if root:
            root = os.path.realpath(root)

        plugins = parse_plugins(args.plugins)
        for p in plugins:
            dr.load_components(p, continue_on_error=False)

        if args.config:
            with open(args.config) as f:
                config = (yaml.safe_load(f))
                packages_loaded = load_packages(config.get('packages', []))
                plugins.extend(packages_loaded)
                apply_default_enabled(config)
                apply_configs(config)

        if component is None:
            component = []
            plugins = tuple(plugins)
            for c in dr.DELEGATES:
                if c.__module__.startswith(plugins):
                    component.append(c)

    if component:
        if not isinstance(component, (list, set)):
            component = [component]

        if args and args.pkg_query:
            pred = taglang.parse(args.pkg_query)
            component = [c for c in component if pred([dr.get_module_name(c)])]
            if not component:
                msg = "No components for pkg-query expression: %s" % args.pkg_query
                raise Exception(msg)

        if args and args.tags:
            pred = taglang.parse(args.tags)
            component = [c for c in component if pred(dr.get_tags(c))]
            if not component:
                msg = "No components for tag expression: %s" % args.tags
                raise Exception(msg)

        graph = {}
        for c in component:
            graph.update(dr.get_dependency_graph(c))
    else:
        graph = dr.COMPONENTS[dr.GROUPS.single]

    broker = dr.Broker()

    if args and args.bare:
        ctx = ExecutionContext()  # dummy context that no spec depend on. needed for filters to work
        specs = parse_specs(args.bare)
        specs = load_specs(specs, ctx)

        broker = dr.Broker()
        broker[ExecutionContext] = ctx
        for spec, content in specs.items():
            broker[spec] = content if dr.DELEGATES[spec].multi_output else content[-1]
    try:
        if formatters:
            for formatter in formatters:
                formatter.preprocess(broker)

            if args and args.bare:
                broker = dr.run(graph, broker=broker)
            else:
                broker = _run(broker, graph, root, context=context, inventory=inventory)

            for formatter in formatters:
                formatter.postprocess(broker)
        elif print_component:
            if args and args.bare:
                broker = dr.run(graph, broker=broker)
            else:
                broker = _run(broker, graph, root, context=context, inventory=inventory)

            broker.print_component(print_component)
        else:
            if args and args.bare:
                broker = dr.run(graph, broker=broker)
            else:
                broker = _run(broker, graph, root, context=context, inventory=inventory)

        return broker
    except (InvalidContentType, InvalidArchive):
        if args and args.archive:
            path = args.archive
            msg = "Invalid directory or archive. Did you mean to pass -p {p}?"
            log.error(msg.format(p=path))
        else:
            raise


def parse_specs(specs):
    """
    -b "hostname=/etc/hostname, redhat_release=/etc/redhat-release, .."
    """
    return dict(s.strip().split("=", 1) for s in specs.split(","))


def load_datasource(name):
    name = "insights.specs.Specs." + name if "." not in name else name
    return dr.get_component(name)


def load_specs(specs, ctx):
    results = defaultdict(list)
    for spec, path in specs.items():
        s = load_datasource(spec)
        root = "/" if path.startswith('/') else "."
        if s.raw:
            c = RawFileProvider(relative_path=path, root=root, ds=s, ctx=ctx)
        else:
            c = TextFileProvider(relative_path=path, root=root, ds=s, ctx=ctx)
        results[s].append(c)
    return results


def main():
    if "" not in sys.path:
        sys.path.insert(0, "")
    run(print_summary=True)


if __name__ == "__main__":
    main()



import doctest
import pytest

from insights.parsers import systemd_analyze
from insights.parsers import SkipException
from insights.tests import context_wrap


OUTPUT = """
33.080s cloud-init-local.service
32.423s unbound-anchor.service
 2.773s kdump.service
 1.699s dnf-makecache.service
 1.304s cloud-init.service
 1.073s initrd-switch-root.service
  939ms cloud-config.service
  872ms tuned.service
  770ms cloud-final.service
""".strip()


def test_output():
    output = systemd_analyze.SystemdAnalyzeBlame(context_wrap(OUTPUT))
    assert ('cloud-init-local.service' in output) is True

    # Test time(seconds)
    assert output.get('cloud-init.service', 0) == 1.304

    with pytest.raises(SkipException):
        assert systemd_analyze.SystemdAnalyzeBlame(context_wrap("")) is None


def test_documentation():
    failed_count, tests = doctest.testmod(
        systemd_analyze,
        globs={'output': systemd_analyze.SystemdAnalyzeBlame(context_wrap(OUTPUT))}
    )
    assert failed_count == 0



"""
multipath_conf parses multipath.conf configuration files into nested
dictionaries.
"""
import string
from insights.parsr import (EOF, Forward, LeftCurly, Lift, Literal, LineEnd,
        RightCurly, Many, Number, OneLineComment, PosMarker, skip_none, String,
        QuotedString, WS, WSChar)
from insights.parsr.query import Entry


def loads(data):
    return Entry(children=Top(data)[0])


def load(f):
    return loads(f.read())


def to_entry(name, rest):
    if isinstance(rest, list):
        return Entry(name=name.value, children=rest, lineno=name.lineno)
    return Entry(name=name.value, attrs=[rest], lineno=name.lineno)


Stmt = Forward()
Num = Number & (WSChar | LineEnd)
NULL = Literal("none", value=None)
Comment = (WS >> OneLineComment("#").map(lambda x: None))
BeginBlock = (WS >> LeftCurly << WS)
EndBlock = (WS >> RightCurly << WS)
Bare = String(set(string.printable) - (set(string.whitespace) | set("#{}'\"")))
Name = WS >> PosMarker(String(string.ascii_letters + "_")) << WS
Value = WS >> (Num | NULL | QuotedString | Bare) << WS
Block = BeginBlock >> Many(Stmt).map(skip_none) << EndBlock
Stanza = (Lift(to_entry) * Name * (Block | Value)) | Comment
Stmt <= WS >> Stanza << WS
Doc = Many(Stmt).map(skip_none)
Top = Doc + EOF



import shlex
import subprocess
from subprocess import PIPE
import six
from threading import Lock

try:
    from insights.contrib import magic
except Exception:
    magic_loaded = False
else:
    # RHEL 6 does not have MAGIC_MIME_TYPE defined, but passing in the value
    # found in RHEL 7 (16, base 10), seems to work.
    mime_flag = magic.MAGIC_MIME_TYPE if hasattr(magic, "MAGIC_MIME_TYPE") else 16
    _magic = magic.open(mime_flag | magic.CONTINUE)
    _magic.load()

    magic_loaded = True


# libmagic is not thread safe so we must lock access to file
magic_lock = Lock()


def from_file(name):
    if magic_loaded:
        with magic_lock:
            return six.b(_magic.file(name)).decode("unicode-escape").splitlines()[0].strip()
    else:
        cmd = "file --mime-type -b %s"
        p = subprocess.Popen(shlex.split(cmd % name), stdout=subprocess.PIPE)
        stdout, _ = p.communicate()
        return stdout.strip().decode("utf-8")


def from_buffer(b):
    if magic_loaded:
        with magic_lock:
            return _magic.buffer(b)
    else:
        cmd = "file --mime-type -b -"
        p = subprocess.Popen(shlex.split(cmd), stdin=PIPE, stdout=PIPE)
        stdout, stderr = p.communicate(b)
        return stdout.strip().decode("utf-8")



from insights.parsers.os_release import OsRelease
from insights.tests import context_wrap

REHL_OS_RELEASE = """
NAME="Red Hat Enterprise Linux Server"
VERSION="7.2 (Maipo)"
ID="rhel"
ID_LIKE="fedora"
VERSION_ID="7.2"
PRETTY_NAME="Employee SKU"
ANSI_COLOR="0;31"
CPE_NAME="cpe:/o:redhat:enterprise_linux:7.2:GA:server"
HOME_URL="https://www.redhat.com/"
BUG_REPORT_URL="https://bugzilla.redhat.com/"

REDHAT_BUGZILLA_PRODUCT="Red Hat Enterprise Linux 7"
REDHAT_BUGZILLA_PRODUCT_VERSION=7.2
REDHAT_SUPPORT_PRODUCT="Red Hat Enterprise Linux"
REDHAT_SUPPORT_PRODUCT_VERSION="7.2"
""".strip()

RHEVH_RHV40_OS_RELEASE = """
NAME="Red Hat Enterprise Linux"
VERSION="7.3"
VERSION_ID="7.3"
ID="rhel"
ID_LIKE="fedora"
VARIANT="Red Hat Virtualization Host"
VARIANT_ID="ovirt-node"
PRETTY_NAME="Red Hat Virtualization Host 4.0 (el7.3)"
ANSI_COLOR="0;31"
CPE_NAME="cpe:/o:redhat:enterprise_linux:7.3:GA:hypervisor"
HOME_URL="https://www.redhat.com/"
BUG_REPORT_URL="https://bugzilla.redhat.com/"

# FIXME
# REDHAT_BUGZILLA_PRODUCT="Red Hat Virtualization"
# REDHAT_BUGZILLA_PRODUCT_VERSION=7.3
# REDHAT_SUPPORT_PRODUCT="Red Hat Virtualization"
# REDHAT_SUPPORT_PRODUCT_VERSION=7.3
""".strip()

FEDORA_OS_RELEASE = """
NAME=Fedora
VERSION="24 (Server Edition)"
ID=fedora
VERSION_ID=24
PRETTY_NAME="Fedora 24 (Server Edition)"
ANSI_COLOR="0;34"
CPE_NAME="cpe:/o:fedoraproject:fedora:24"
HOME_URL="https://fedoraproject.org/"
BUG_REPORT_URL="https://bugzilla.redhat.com/"
REDHAT_BUGZILLA_PRODUCT="Fedora"
REDHAT_BUGZILLA_PRODUCT_VERSION=24
REDHAT_SUPPORT_PRODUCT="Fedora"
REDHAT_SUPPORT_PRODUCT_VERSION=24
PRIVACY_POLICY_URL=https://fedoraproject.org/wiki/Legal:PrivacyPolicy
VARIANT="Server Edition"
VARIANT_ID=server
""".strip()


def test_rhel():
    rls = OsRelease(context_wrap(REHL_OS_RELEASE))
    data = rls.data
    assert data.get("VARIANT_ID") is None
    assert data.get("VERSION") == "7.2 (Maipo)"


def test_rhevh():
    rls = OsRelease(context_wrap(RHEVH_RHV40_OS_RELEASE))
    data = rls.data
    assert data.get("VARIANT_ID") == "ovirt-node"
    assert data.get("VERSION") == "7.3"
    assert data.get("PRETTY_NAME") == "Red Hat Virtualization Host 4.0 (el7.3)"


def test_fedora24():
    rls = OsRelease(context_wrap(FEDORA_OS_RELEASE))
    data = rls.data
    assert data.get("VARIANT_ID") == "server"
    assert data.get("VERSION") == "24 (Server Edition)"
    assert data.get("PRETTY_NAME") == "Fedora 24 (Server Edition)"



from insights.tests import context_wrap
from insights.parsers.rhn_logs import TaskomaticDaemonLog, SearchDaemonLog
from insights.parsers.rhn_logs import ServerXMLRPCLog
from insights.parsers.rhn_logs import SatelliteServerLog
from datetime import datetime

search_daemon_log = """
STATUS | wrapper  | 2013/01/28 14:22:38 | TERM trapped.  Shutting down.
STATUS | wrapper  | 2013/01/28 14:22:39 | <-- Wrapper Stopped
STATUS | wrapper  | 2013/01/28 14:41:58 | --> Wrapper Started as Daemon
STATUS | wrapper  | 2013/01/28 14:41:58 | Launching a JVM...
INFO   | jvm 1    | 2013/01/28 14:41:59 | Wrapper (Version 3.2.1) http://wrapper.tanukisoftware.org
STATUS | wrapper  | 2013/01/29 17:04:25 | TERM trapped.  Shutting down.
""".strip()

taskomatic_daemon_log = """
STATUS | wrapper  | 2016/05/18 15:13:36 | --> Wrapper Started as Daemon
STATUS | wrapper  | 2016/05/18 15:13:36 | Launching a JVM...
INFO   | jvm 1    | 2016/05/18 15:13:36 | Wrapper (Version 3.2.3) http://wrapper.tanukisoftware.org
INFO   | jvm 1    | 2016/05/18 15:13:36 |   Copyright 1999-2006 Tanuki Software, Inc.  All Rights Reserved.
INFO   | jvm 1    | 2016/05/18 15:13:39 | May 18, 2016 3:13:39 PM com.mchange.v2.log.MLog <clinit>
INFO   | jvm 1    | 2016/05/18 15:13:39 | INFO: MLog clients using java 1.4+ standard logging.
INFO   | jvm 1    | 2016/05/18 15:13:39 | May 18, 2016 3:13:39 PM com.mchange.v2.c3p0.C3P0Registry banner
INFO   | jvm 1    | 2016/05/18 15:13:39 | INFO: Initializing c3p0-0.9.1.2 [built 31-March-2011 15:45:28; debug? false; trace: 5]
INFO   | jvm 1    | 2016/05/18 15:13:39 | May 18, 2016 3:13:39 PM com.mchange.v2.c3p0.impl.AbstractPoolBackedDataSource getPoolManager
INFO   | jvm 1    | 2016/05/18 15:13:40 | INFO: Initializing c3p0 pool... com.mchange.v2.c3p0.PoolBackedDataSource@ea9d5b40
INFO   | jvm 1
""".strip()

taskomatic_daemon_bad_date = """
INFO   | jvm 1    | rbad/da/ta 15:13:40 | INFO: Initializing c3p0 pool... com.mchange.v2.c3p0.PoolBackedDataSource@ea9d5b40
"""


def test_rhn_search_daemon_log():
    out_log = SearchDaemonLog(context_wrap(search_daemon_log))
    assert "Wrapper Started as Daemon" in out_log
    assert len(out_log.get("jvm")) == 1
    assert out_log.get("jvm")[0]['raw_message'] == 'INFO   | jvm 1    | 2013/01/28 14:41:59 | Wrapper (Version 3.2.1) http://wrapper.tanukisoftware.org'
    assert len(list(out_log.get_after(datetime(2013, 1, 28, 19, 30, 0)))) == 1


def test_rhn_taskomatic_daemon_log():
    out_log = TaskomaticDaemonLog(context_wrap(taskomatic_daemon_log))
    assert "Wrapper Started as Daemon" in out_log
    assert len(out_log.get("jvm")) == 9
    assert out_log.get("jvm")[2]['raw_message'] == 'INFO   | jvm 1    | 2016/05/18 15:13:39 | May 18, 2016 3:13:39 PM com.mchange.v2.log.MLog <clinit>'
    assert out_log.last_log_date == datetime(2016, 5, 18, 15, 13, 40)
    assert len(list(out_log.get_after(datetime(2016, 5, 18, 15, 13, 38)))) == 7


def test_bad_date_handling():
    out_log = TaskomaticDaemonLog(context_wrap(taskomatic_daemon_bad_date))
    assert 'Initializing c3p0 pool..' in out_log
    assert out_log.last_log_date is None


SERVER_XMLRPC_LOG_DATA = """
2016/04/11 05:52:01 -04:00 23630 10.4.4.17: xmlrpc/registration.welcome_message('lang: None',)
2016/04/11 05:52:26 -04:00 12911 10.4.4.17: xmlrpc/registration.create_system("token = '1-RegKey'", '6Server', 'x86_64')
2016/04/11 05:52:26 -04:00 12911 10.4.4.17: rhnServer/server_token.process_token(1000010125, 'enterprise_entitled,provisioning_entitled,virtualization_host')
2016/04/11 05:52:47 -04:00 23628 10.4.4.17: xmlrpc/registration.__add_hw_profile_no_auth(1000010125, 'items: 6')
2016/04/11 10:49:12 -04:00 11594 192.168.18.163: xmlrpc/registration.welcome_message('lang: None',)
2016/04/11 10:49:12 -04:00 11601 192.168.18.163: xmlrpc/registration.register_osad
2016/04/11 10:49:17 -04:00 11599 192.168.18.163: xmlrpc/registration.register_osad_jid
2016/04/11 12:57:26 -04:00 11594 192.168.18.28: rhnServer/server_certificate.valid('Server id ID-1000010124 not found in database',)
2016/07/27 04:44:41 -04:00 25923 2620:10a:0:4::40: xmlrpc/queue.get(1000014812, 2, 'checkins enabled')
"""


def test_server_xmlrpc_log_data():
    log = ServerXMLRPCLog(context_wrap(SERVER_XMLRPC_LOG_DATA))

    # Check the first log line for all fields
    line = log.get('10.4.4.17: xmlrpc/registration.welcome_message')[0]

    assert line['timestamp'] == '2016/04/11 05:52:01 -04:00'
    d = datetime(2016, 0o4, 11, 0o5, 52, 0o1)
    assert line['datetime'] == d
    assert line['pid'] == '23630'
    assert line['client_ip'] == '10.4.4.17'
    assert line['module'] == 'xmlrpc'
    assert line['function'] == 'registration.welcome_message'
    assert line['client_id'] is None
    assert line['args'] == "'lang: None'"

    # Check that get works
    assert len(log.get('welcome_message')) == 2

    # Test get_after
    assert len(list(log.get_after(datetime(2016, 7, 27, 0, 0, 0)))) == 1

    # Check that __contains__ works
    assert 'welcome_message' in log

    # Check parsing of lines without argument lists
    line = log.get('registration.register_osad_jid')[0]
    assert line['args'] is None

    # Check lines that include the client ID
    line = log.get('__add_hw_profile_no_auth')[0]
    assert line['client_id'] == '1000010125'
    assert line['args'] == "'items: 6'"

    # Check that we can get IPv6 addresses correctly
    line = log.get('checkins enabled')[0]
    assert line['client_ip'] == '2620:10a:0:4::40'
    assert line['client_id'] == '1000014812'
    assert line['args'] == "2, 'checkins enabled'"
    # Use this line for the last test later

    # Test last attribute
    last = log.last
    assert line == last


SERVER_XMLRPC_LOG_BAD_DATE = """
2016/13/35 05:52:01 -04:00 23630 10.4.4.17: xmlrpc/registration.welcome_message('lang: None',)
Received SIGHUP
""".strip()

SERVER_XMLRPC_LOG_NO_DATE = """
/usr/bin/httpd: command or file not found
""".strip()


def test_server_xmlrpc_log_bad_date():
    log = ServerXMLRPCLog(context_wrap(SERVER_XMLRPC_LOG_BAD_DATE))
    assert len(log.get('registration.welcome_message')) == 1
    line = log.get('registration.welcome_message')[0]
    assert line['timestamp'] == '2016/13/35 05:52:01 -04:00'
    assert 'datetime' not in line
    log = ServerXMLRPCLog(context_wrap(SERVER_XMLRPC_LOG_NO_DATE))
    assert len(log.get('registration.welcome_message')) == 0
    assert log.last == {'raw_message': '/usr/bin/httpd: command or file not found'}


RHN_SERVER_SATELLITE_LOG = """
2016/11/19 01:13:35 -04:00    Retrieving / parsing errata data: rhel-x86_64-server-optional-6-debuginfo (0)
2016/11/19 01:13:35 -04:00 Downloading errata data complete
2016/11/19 01:13:35 -04:00
2016/11/19 01:13:35 -04:00 Downloading kickstartable trees metadata
2016/11/19 01:13:35 -04:00    Retrieving / parsing kickstart data: rhel-x86_64-server-optional-6-debuginfo (NONE RELEVANT)
2016/11/19 01:13:35 -04:00
2016/11/19 01:13:35 -04:00 Downloading kickstartable trees files
2016/11/19 01:13:35 -04:00    Retrieving / parsing kickstart tree files: rhel-x86_64-server-optional-6-debuginfo (NONE RELEVANT)
2016/11/19 01:13:44 -04:00 channel-families data complete
2016/11/19 01:13:44 -04:00
2016/11/19 01:13:44 -04:00 RHN Entitlement Certificate sync
"""


def test_rhn_server_satellite_get():
    log = SatelliteServerLog(context_wrap(RHN_SERVER_SATELLITE_LOG))
    assert len(log.get('Downloading')) == 3
    assert len(list(log.get_after(datetime(2016, 11, 19, 1, 13, 44)))) == 3



from insights.parsers.tuned import Tuned
from insights.tests import context_wrap

TUNED_OUTPUT = '''
Available profiles:
- balanced
- desktop
- latency-performance
- network-latency
- network-throughput
- powersave
- throughput-performance
- virtual-guest
- virtual-host
Current active profile: virtual-guest
'''.strip()

TUNED_OUTPUT2 = '''
Available profiles:
- balanced
- desktop
- latency-performance
- network-latency
- network-throughput
- powersave
- throughput-performance
- virtual-guest
- virtual-host
It seems that tuned daemon is not running, preset profile is not activated.
Preset profile: virtual-guest
'''.strip()


def test_active_profile():
    tuned_output = Tuned(context_wrap(TUNED_OUTPUT))
    assert len(tuned_output.data.get('available')) == 9
    assert tuned_output.data.get('active') == 'virtual-guest'
    assert tuned_output.data.get('available') == ['balanced',
                                                  'desktop',
                                                  'latency-performance',
                                                  'network-latency',
                                                  'network-throughput',
                                                  'powersave',
                                                  'throughput-performance',
                                                  'virtual-guest',
                                                  'virtual-host']


def test_preset_profile():
    tuned_output = Tuned(context_wrap(TUNED_OUTPUT2))
    assert len(tuned_output.data.get('available')) == 9
    assert tuned_output.data.get('preset') == 'virtual-guest'
    assert tuned_output.data.get('available') == ['balanced',
                                                  'desktop',
                                                  'latency-performance',
                                                  'network-latency',
                                                  'network-throughput',
                                                  'powersave',
                                                  'throughput-performance',
                                                  'virtual-guest',
                                                  'virtual-host']



from .. import Parser, parser, get_active_lines, LegacyItemAccess
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.up2date)
class Up2Date(LegacyItemAccess, Parser):
    """
    .. warning::
        This parser is deprecated, please use
        :py:class:`insights.parsers.sysconfig.Up2DateSysconfig` instead.

    Class to parse the ``up2date``

    Attributes:
        data (dict): A dict of up2date info which ignores comment lines.
        The first and second line for key word 'serverURL' will be ignored.

    For example:
        serverURL[comment]=Remote server URL
        #serverURL=https://rhnproxy.glb.tech.markit.partners/XMLRPC
        serverURL=https://rhnproxy.glb.tech.markit.partners/XMLRPC
    """
    def __init__(self, *args, **kwargs):
        deprecated(Up2Date, "Import Up2DateSysconfig from insights.parsers.sysconfig instead")
        super(Up2Date, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        up2date_info = {}
        for line in get_active_lines(content):
            if "[comment]" not in line and '=' in line:
                key, val = line.split('=')
                up2date_info[key.strip()] = val.strip()
        self.data = up2date_info



# -*- coding: UTF-8 -*-

"""
AbrtStatusBare - command ``/usr/bin/abrt status --bare=True``
=============================================================

``/usr/bin/abrt status --bare=True`` returns the number of problems ABRT
detected in the system.

Examples:
    >>> abrt_status_bare.problem_count
    1997
"""

from insights import CommandParser, parser
from insights.specs import Specs


@parser(Specs.abrt_status_bare)
class AbrtStatusBare(CommandParser):
    """
    Parser for the output of ``abrt status --bare=True``

    Attributes:
        problem_count (int): the number of problems ABRT detected
    """
    def parse_content(self, content):
        self.problem_count = int(content[0].strip())



"""
This module defines all datasources used by standard Red Hat Insight components.

To define data sources that override the components in this file, create a
`insights.core.spec_factory.SpecFactory` with "insights.specs" as the constructor
argument. Data sources created with that factory will override components in
this file with the same `name` keyword argument. This allows overriding the
data sources that standard Insights `Parsers` resolve against.
"""

import logging
import os
import re

from insights.core.context import HostContext

from insights.core.dr import SkipComponent
from insights.core.plugins import datasource
from insights.core.spec_factory import RawFileProvider
from insights.core.spec_factory import simple_file, simple_command, glob_file
from insights.core.spec_factory import first_of, foreach_collect, foreach_execute
from insights.core.spec_factory import first_file, listdir
from insights.combiners.cloud_provider import CloudProvider
from insights.combiners.services import Services
from insights.combiners.sap import Sap
from insights.specs import Specs


from grp import getgrgid
from os import stat
from pwd import getpwuid


logger = logging.getLogger(__name__)


def get_owner(filename):
    """ tuple: Return tuple containing uid and gid of file filename """
    st = stat(filename)
    name = getpwuid(st.st_uid).pw_name
    group = getgrgid(st.st_gid).gr_name
    return (name, group)


def get_cmd_and_package_in_ps(broker, target_command):
    """
    Search for command in ``ps auxww`` output and determine RPM providing binary

    Arguments:
        broker(dict): Current state of specs collected by Insights
        target_command(str): Command name to search for in ps output

    Returns:
        set: Set including all RPMs that provide the target command
    """
    ps = broker[DefaultSpecs.ps_auxww].content
    ctx = broker[HostContext]
    results = set()
    for p in ps:
        p_splits = p.split(None, 10)
        cmd = p_splits[10].split()[0] if len(p_splits) == 11 else ''
        which = ctx.shell_out("which {0}".format(cmd)) if target_command in os.path.basename(cmd) else None
        resolved = ctx.shell_out("readlink -e {0}".format(which[0])) if which else None
        pkg = ctx.shell_out("/bin/rpm -qf {0}".format(resolved[0])) if resolved else None
        if cmd and pkg is not None:
            results.add("{0} {1}".format(cmd, pkg[0]))
    return results


def _make_rpm_formatter(fmt=None):
    """ function: Returns function that will format output of rpm query command """
    if fmt is None:
        fmt = [
            '"name":"%{NAME}"',
            '"epoch":"%{EPOCH}"',
            '"version":"%{VERSION}"',
            '"release":"%{RELEASE}"',
            '"arch":"%{ARCH}"',
            '"installtime":"%{INSTALLTIME:date}"',
            '"buildtime":"%{BUILDTIME}"',
            '"vendor":"%{VENDOR}"',
            '"buildhost":"%{BUILDHOST}"',
            '"sigpgp":"%{SIGPGP:pgpsig}"'
        ]

    def inner(idx=None):
        if idx:
            return "\{" + ",".join(fmt[:idx]) + "\}\n"
        else:
            return "\{" + ",".join(fmt) + "\}\n"
    return inner


format_rpm = _make_rpm_formatter()


class DefaultSpecs(Specs):
    abrt_ccpp_conf = simple_file("/etc/abrt/plugins/CCpp.conf")
    abrt_status_bare = simple_command("/usr/bin/abrt status --bare=True")
    amq_broker = glob_file("/var/opt/amq-broker/*/etc/broker.xml")
    auditctl_status = simple_command("/sbin/auditctl -s")
    auditd_conf = simple_file("/etc/audit/auditd.conf")
    audit_log = simple_file("/var/log/audit/audit.log")
    avc_hash_stats = simple_file("/sys/fs/selinux/avc/hash_stats")
    avc_cache_threshold = simple_file("/sys/fs/selinux/avc/cache_threshold")

    @datasource(CloudProvider)
    def is_aws(broker):
        """ bool: Returns True if this node is identified as running in AWS """
        cp = broker[CloudProvider]
        if cp and cp.cloud_provider == CloudProvider.AWS:
            return True
        raise SkipComponent()

    aws_instance_id_doc = simple_command("/usr/bin/curl -s http://169.254.169.254/latest/dynamic/instance-identity/document --connect-timeout 5", deps=[is_aws])
    aws_instance_id_pkcs7 = simple_command("/usr/bin/curl -s http://169.254.169.254/latest/dynamic/instance-identity/pkcs7 --connect-timeout 5", deps=[is_aws])
    awx_manage_check_license = simple_command("/usr/bin/awx-manage check_license")

    @datasource(CloudProvider)
    def is_azure(broker):
        """ bool: Returns True if this node is identified as running in Azure """
        cp = broker[CloudProvider]
        if cp and cp.cloud_provider == CloudProvider.AZURE:
            return True
        raise SkipComponent()

    azure_instance_type = simple_command("/usr/bin/curl -s -H Metadata:true http://169.254.169.254/metadata/instance/compute/vmSize?api-version=2018-10-01&format=text --connect-timeout 5", deps=[is_azure])
    bios_uuid = simple_command("/usr/sbin/dmidecode -s system-uuid")
    blkid = simple_command("/sbin/blkid -c /dev/null")
    bond = glob_file("/proc/net/bonding/bond*")
    bond_dynamic_lb = glob_file("/sys/class/net/bond[0-9]*/bonding/tlb_dynamic_lb")
    boot_loader_entries = glob_file("/boot/loader/entries/*.conf")
    branch_info = simple_file("/branch_info", kind=RawFileProvider)
    brctl_show = simple_command("/usr/sbin/brctl show")
    candlepin_log = simple_file("/var/log/candlepin/candlepin.log")
    cgroups = simple_file("/proc/cgroups")
    ps_alxwww = simple_command("/bin/ps alxwww")
    ps_aux = simple_command("/bin/ps aux")
    ps_auxcww = simple_command("/bin/ps auxcww")
    ps_auxww = simple_command("/bin/ps auxww")
    ps_ef = simple_command("/bin/ps -ef")
    ps_eo = simple_command("/usr/bin/ps -eo pid,ppid,comm")

    @datasource(ps_auxww)
    def tomcat_base(broker):
        """
        Function to search the output of ``ps auxww`` to find all running tomcat
        processes and extract the base path where the process was started.

        Returns:
            list: List of the paths to each running process
        """
        ps = broker[DefaultSpecs.ps_auxww].content
        results = []
        findall = re.compile(r"\-Dcatalina\.base=(\S+)").findall
        for p in ps:
            found = findall(p)
            if found:
                # Only get the path which is absolute
                results.extend(f for f in found if f[0] == '/')
        return list(set(results))

    catalina_out = foreach_collect(tomcat_base, "%s/catalina.out")
    cciss = glob_file("/proc/driver/cciss/cciss*")
    cdc_wdm = simple_file("/sys/bus/usb/drivers/cdc_wdm/module/refcnt")
    ceilometer_collector_log = first_file(["/var/log/containers/ceilometer/collector.log", "/var/log/ceilometer/collector.log"])
    ceilometer_compute_log = first_file(["/var/log/containers/ceilometer/compute.log", "/var/log/ceilometer/compute.log"])
    ceilometer_conf = first_file(["/var/lib/config-data/puppet-generated/ceilometer/etc/ceilometer/ceilometer.conf", "/etc/ceilometer/ceilometer.conf"])
    ceph_conf = first_file(["/var/lib/config-data/puppet-generated/ceph/etc/ceph/ceph.conf", "/etc/ceph/ceph.conf"])
    ceph_df_detail = simple_command("/usr/bin/ceph df detail -f json")
    ceph_health_detail = simple_command("/usr/bin/ceph health detail -f json")

    @datasource(ps_auxww)
    def is_ceph_monitor(broker):
        """ bool: Returns True if ceph monitor process ceph-mon is running on this node """
        ps = broker[DefaultSpecs.ps_auxww].content
        findall = re.compile(r"ceph\-mon").findall
        if any(findall(p) for p in ps):
            return True
        raise SkipComponent()

    ceph_insights = simple_command("/usr/bin/ceph insights", deps=[is_ceph_monitor])
    ceph_log = glob_file(r"var/log/ceph/ceph.log*")
    ceph_osd_dump = simple_command("/usr/bin/ceph osd dump -f json")
    ceph_osd_ec_profile_ls = simple_command("/usr/bin/ceph osd erasure-code-profile ls")
    ceph_osd_log = glob_file(r"var/log/ceph/ceph-osd*.log")
    ceph_osd_tree = simple_command("/usr/bin/ceph osd tree -f json")
    ceph_s = simple_command("/usr/bin/ceph -s -f json")
    ceph_v = simple_command("/usr/bin/ceph -v")
    certificates_enddate = simple_command("/usr/bin/find /etc/origin/node /etc/origin/master /etc/pki /etc/ipa -type f -exec /usr/bin/openssl x509 -noout -enddate -in '{}' \; -exec echo 'FileName= {}' \;")
    chkconfig = simple_command("/sbin/chkconfig --list")
    chrony_conf = simple_file("/etc/chrony.conf")
    chronyc_sources = simple_command("/usr/bin/chronyc sources")
    cib_xml = simple_file("/var/lib/pacemaker/cib/cib.xml")
    cinder_api_log = first_file(["/var/log/containers/cinder/cinder-api.log", "/var/log/cinder/cinder-api.log"])
    cinder_conf = first_file(["/var/lib/config-data/puppet-generated/cinder/etc/cinder/cinder.conf", "/etc/cinder/cinder.conf"])
    cinder_volume_log = first_file(["/var/log/containers/cinder/volume.log", "/var/log/containers/cinder/cinder-volume.log", "/var/log/cinder/volume.log"])
    cloud_init_custom_network = simple_file("/etc/cloud/cloud.cfg.d/99-custom-networking.cfg")
    cloud_init_log = simple_file("/var/log/cloud-init.log")
    cluster_conf = simple_file("/etc/cluster/cluster.conf")
    cmdline = simple_file("/proc/cmdline")
    cobbler_settings = first_file(["/etc/cobbler/settings", "/conf/cobbler/settings"])
    cobbler_modules_conf = first_file(["/etc/cobbler/modules.conf", "/conf/cobbler/modules.conf"])
    corosync = simple_file("/etc/sysconfig/corosync")
    corosync_conf = simple_file("/etc/corosync/corosync.conf")
    cpu_cores = glob_file("sys/devices/system/cpu/cpu[0-9]*/online")
    cpu_siblings = glob_file("sys/devices/system/cpu/cpu[0-9]*/topology/thread_siblings_list")
    cpu_smt_active = simple_file("sys/devices/system/cpu/smt/active")
    cpu_vulns = glob_file("sys/devices/system/cpu/vulnerabilities/*")
    cpuinfo = simple_file("/proc/cpuinfo")
    cpupower_frequency_info = simple_command("/usr/bin/cpupower -c all frequency-info")
    cpuset_cpus = simple_file("/sys/fs/cgroup/cpuset/cpuset.cpus")
    cron_daily_rhsmd = simple_file("/etc/cron.daily/rhsmd")
    crypto_policies_config = simple_file("/etc/crypto-policies/config")
    crypto_policies_state_current = simple_file("/etc/crypto-policies/state/current")
    crypto_policies_opensshserver = simple_file("/etc/crypto-policies/back-ends/opensshserver.config")
    crypto_policies_bind = simple_file("/etc/crypto-policies/back-ends/bind.config")
    current_clocksource = simple_file("/sys/devices/system/clocksource/clocksource0/current_clocksource")
    date = simple_command("/bin/date")
    date_utc = simple_command("/bin/date --utc")
    designate_conf = first_file(["/var/lib/config-data/puppet-generated/designate/etc/designate/designate.conf",
                                 "/etc/designate/designate.conf"])
    df__al = simple_command("/bin/df -al -x autofs")
    df__alP = simple_command("/bin/df -alP -x autofs")
    df__li = simple_command("/bin/df -li -x autofs")
    dig_dnssec = simple_command("/usr/bin/dig +dnssec . SOA")
    dig_edns = simple_command("/usr/bin/dig +edns=0 . SOA")
    dig_noedns = simple_command("/usr/bin/dig +noedns . SOA")
    dirsrv_errors = glob_file("var/log/dirsrv/*/errors*")
    dm_mod_use_blk_mq = simple_file("/sys/module/dm_mod/parameters/use_blk_mq")
    dmesg = simple_command("/bin/dmesg")
    dmesg_log = simple_file("/var/log/dmesg")
    dmidecode = simple_command("/usr/sbin/dmidecode")
    dmsetup_info = simple_command("/usr/sbin/dmsetup info -C")
    dnf_conf = simple_file("/etc/dnf/dnf.conf")
    dnf_modules = glob_file("/etc/dnf/modules.d/*.module")
    docker_info = simple_command("/usr/bin/docker info")
    docker_list_containers = simple_command("/usr/bin/docker ps --all --no-trunc")
    docker_list_images = simple_command("/usr/bin/docker images --all --no-trunc --digests")
    doveconf = simple_command("/usr/bin/doveconf")
    docker_storage_setup = simple_file("/etc/sysconfig/docker-storage-setup")
    docker_sysconfig = simple_file("/etc/sysconfig/docker")
    dotnet_version = simple_command("/usr/bin/dotnet --version")
    dracut_kdump_capture_service = simple_file("/usr/lib/dracut/modules.d/99kdumpbase/kdump-capture.service")

    @datasource()
    def du_dirs_list(broker):
        """ Provide a list of directorys for the ``du_dirs`` spec to scan """
        return ['/var/lib/candlepin/activemq-artemis']
    du_dirs = foreach_execute(du_dirs_list, "/bin/du -s -k %s")
    engine_db_query_vdsm_version = simple_command('engine-db-query --statement "SELECT vs.vds_name, rpm_version FROM vds_dynamic vd, vds_static vs WHERE vd.vds_id = vs.vds_id" --json')
    engine_log = simple_file("/var/log/ovirt-engine/engine.log")
    etc_journald_conf = simple_file(r"etc/systemd/journald.conf")
    etc_journald_conf_d = glob_file(r"etc/systemd/journald.conf.d/*.conf")
    etc_machine_id = simple_file("/etc/machine-id")
    etc_udev_40_redhat_rules = simple_file("/etc/udev/rules.d/40-redhat.rules")
    etcd_conf = simple_file("/etc/etcd/etcd.conf")
    ethernet_interfaces = listdir("/sys/class/net", context=HostContext)
    ethtool = foreach_execute(ethernet_interfaces, "/sbin/ethtool %s")
    ethtool_S = foreach_execute(ethernet_interfaces, "/sbin/ethtool -S %s")
    ethtool_T = foreach_execute(ethernet_interfaces, "/sbin/ethtool -T %s")
    ethtool_c = foreach_execute(ethernet_interfaces, "/sbin/ethtool -c %s")
    ethtool_g = foreach_execute(ethernet_interfaces, "/sbin/ethtool -g %s")
    ethtool_i = foreach_execute(ethernet_interfaces, "/sbin/ethtool -i %s")
    ethtool_k = foreach_execute(ethernet_interfaces, "/sbin/ethtool -k %s")
    facter = simple_command("/usr/bin/facter")
    fc_match = simple_command("/bin/fc-match -sv 'sans:regular:roman' family fontformat")
    fcoeadm_i = simple_command("/usr/sbin/fcoeadm -i")
    findmnt_lo_propagation = simple_command("/bin/findmnt -lo+PROPAGATION")
    firewall_cmd_list_all_zones = simple_command("/usr/bin/firewall-cmd --list-all-zones")
    firewalld_conf = simple_file("/etc/firewalld/firewalld.conf")
    fstab = simple_file("/etc/fstab")
    galera_cnf = first_file(["/var/lib/config-data/puppet-generated/mysql/etc/my.cnf.d/galera.cnf", "/etc/my.cnf.d/galera.cnf"])
    getconf_page_size = simple_command("/usr/bin/getconf PAGE_SIZE")
    getenforce = simple_command("/usr/sbin/getenforce")
    getsebool = simple_command("/usr/sbin/getsebool -a")
    gluster_v_info = simple_command("/usr/sbin/gluster volume info")
    gnocchi_conf = first_file(["/var/lib/config-data/puppet-generated/gnocchi/etc/gnocchi/gnocchi.conf", "/etc/gnocchi/gnocchi.conf"])
    gnocchi_metricd_log = first_file(["/var/log/containers/gnocchi/gnocchi-metricd.log", "/var/log/gnocchi/metricd.log"])
    grub_conf = simple_file("/boot/grub/grub.conf")
    grub_config_perms = simple_command("/bin/ls -l /boot/grub2/grub.cfg")  # only RHEL7 and updwards
    grub_efi_conf = simple_file("/boot/efi/EFI/redhat/grub.conf")
    grub1_config_perms = simple_command("/bin/ls -l /boot/grub/grub.conf")  # RHEL6
    grub2_cfg = simple_file("/boot/grub2/grub.cfg")
    grub2_efi_cfg = simple_file("boot/efi/EFI/redhat/grub.cfg")
    grubby_default_index = simple_command("/usr/sbin/grubby --default-index")  # only RHEL7 and updwards
    grubby_default_kernel = simple_command("/sbin/grubby --default-kernel")
    hammer_task_list = simple_command("/usr/bin/hammer --config /root/.hammer/cli.modules.d/foreman.yml --output csv task list --search 'state=running AND ( label=Actions::Candlepin::ListenOnCandlepinEvents OR label=Actions::Katello::EventQueue::Monitor )'")
    haproxy_cfg = first_file(["/var/lib/config-data/puppet-generated/haproxy/etc/haproxy/haproxy.cfg", "/etc/haproxy/haproxy.cfg"])
    heat_api_log = first_file(["/var/log/containers/heat/heat_api.log", "/var/log/heat/heat-api.log", "/var/log/heat/heat_api.log"])
    heat_conf = first_file(["/var/lib/config-data/puppet-generated/heat/etc/heat/heat.conf", "/etc/heat/heat.conf"])
    hostname = simple_command("/bin/hostname -f")
    hostname_default = simple_command("/bin/hostname")
    hostname_short = simple_command("/bin/hostname -s")
    hosts = simple_file("/etc/hosts")
    httpd_conf = glob_file(
        [
            "/etc/httpd/conf/httpd.conf",
            "/etc/httpd/conf.d/*.conf",
            "/etc/httpd/conf.d/*/*.conf",
            "/etc/httpd/conf.modules.d/*.conf"
        ]
    )
    httpd_conf_scl_httpd24 = glob_file(
        [
            "/opt/rh/httpd24/root/etc/httpd/conf/httpd.conf",
            "/opt/rh/httpd24/root/etc/httpd/conf.d/*.conf",
            "/opt/rh/httpd24/root/etc/httpd/conf.d/*/*.conf",
            "/opt/rh/httpd24/root/etc/httpd/conf.modules.d/*.conf"
        ]
    )
    httpd_conf_scl_jbcs_httpd24 = glob_file(
        [
            "/opt/rh/jbcs-httpd24/root/etc/httpd/conf/httpd.conf",
            "/opt/rh/jbcs-httpd24/root/etc/httpd/conf.d/*.conf",
            "/opt/rh/jbcs-httpd24/root/etc/httpd/conf.d/*/*.conf",
            "/opt/rh/jbcs-httpd24/root/etc/httpd/conf.modules.d/*.conf"
        ]
    )
    httpd_error_log = simple_file("var/log/httpd/error_log")
    httpd24_httpd_error_log = simple_file("/opt/rh/httpd24/root/etc/httpd/logs/error_log")
    jbcs_httpd24_httpd_error_log = simple_file("/opt/rh/jbcs-httpd24/root/etc/httpd/logs/error_log")
    virt_uuid_facts = simple_file("/etc/rhsm/facts/virt_uuid.facts")

    @datasource(ps_auxww)
    def httpd_cmd(broker):
        """
        Function to search the output of ``ps auxww`` to find all running Apache
        webserver processes and extract the binary path.

        Returns:
            list: List of the binary paths to each running process
        """
        ps = broker[DefaultSpecs.ps_auxww].content
        ps_httpds = set()
        for p in ps:
            p_splits = p.split(None, 10)
            if len(p_splits) >= 11:
                cmd = p_splits[10].split()[0]
                # Should compatible with RHEL6
                # e.g. /usr/sbin/httpd, /usr/sbin/httpd.worker and /usr/sbin/httpd.event
                #      and SCL's httpd24-httpd
                if os.path.basename(cmd).startswith('httpd'):
                    ps_httpds.add(cmd)
        # Running multiple httpd instances on RHEL is supported
        # https://access.redhat.com/solutions/21680
        return list(ps_httpds)

    httpd_M = foreach_execute(httpd_cmd, "%s -M")
    httpd_V = foreach_execute(httpd_cmd, "%s -V")
    ifcfg = glob_file("/etc/sysconfig/network-scripts/ifcfg-*")
    ifcfg_static_route = glob_file("/etc/sysconfig/network-scripts/route-*")
    imagemagick_policy = glob_file(["/etc/ImageMagick/policy.xml", "/usr/lib*/ImageMagick-6.5.4/config/policy.xml"])
    initctl_lst = simple_command("/sbin/initctl --system list")
    init_process_cgroup = simple_file("/proc/1/cgroup")
    interrupts = simple_file("/proc/interrupts")
    ip_addr = simple_command("/sbin/ip addr")
    ip_addresses = simple_command("/bin/hostname -I")
    ip_route_show_table_all = simple_command("/sbin/ip route show table all")
    ip_s_link = simple_command("/sbin/ip -s -d link")
    ipaupgrade_log = simple_file("/var/log/ipaupgrade.log")
    ipcs_m = simple_command("/usr/bin/ipcs -m")
    ipcs_m_p = simple_command("/usr/bin/ipcs -m -p")
    ipcs_s = simple_command("/usr/bin/ipcs -s")
    ipsec_conf = simple_file("/etc/ipsec.conf")
    iptables = simple_command("/sbin/iptables-save")
    iptables_permanent = simple_file("etc/sysconfig/iptables")
    ip6tables = simple_command("/sbin/ip6tables-save")
    ipv4_neigh = simple_command("/sbin/ip -4 neighbor show nud all")
    ipv6_neigh = simple_command("/sbin/ip -6 neighbor show nud all")
    ironic_inspector_log = first_file(["/var/log/containers/ironic-inspector/ironic-inspector.log", "/var/log/ironic-inspector/ironic-inspector.log"])
    iscsiadm_m_session = simple_command("/usr/sbin/iscsiadm -m session")
    kdump_conf = simple_file("/etc/kdump.conf")
    kernel_config = glob_file("/boot/config-*")
    kexec_crash_size = simple_file("/sys/kernel/kexec_crash_size")
    keystone_crontab = simple_command("/usr/bin/crontab -l -u keystone")
    kpatch_list = simple_command("/usr/sbin/kpatch list")
    krb5 = glob_file([r"etc/krb5.conf", r"etc/krb5.conf.d/*"])
    ksmstate = simple_file("/sys/kernel/mm/ksm/run")
    kubepods_cpu_quota = glob_file("/sys/fs/cgroup/cpu/kubepods.slice/kubepods-burstable.slice/kubepods-burstable-pod[a-f0-9_]*.slice/cpu.cfs_quota_us")
    last_upload_globs = ["/etc/redhat-access-insights/.lastupload", "/etc/insights-client/.lastupload"]
    lastupload = glob_file(last_upload_globs)
    libssh_client_config = simple_file("/etc/libssh/libssh_client.config")
    libssh_server_config = simple_file("/etc/libssh/libssh_server.config")
    libvirtd_log = simple_file("/var/log/libvirt/libvirtd.log")
    limits_conf = glob_file(["/etc/security/limits.conf", "/etc/security/limits.d/*.conf"])
    localtime = simple_command("/usr/bin/file -L /etc/localtime")
    logrotate_conf = glob_file(["/etc/logrotate.conf", "/etc/logrotate.d/*"])
    lpstat_p = simple_command("/usr/bin/lpstat -p")
    ls_boot = simple_command("/bin/ls -lanR /boot")
    ls_dev = simple_command("/bin/ls -lanR /dev")
    ls_disk = simple_command("/bin/ls -lanR /dev/disk")
    ls_edac_mc = simple_command("/bin/ls -lan /sys/devices/system/edac/mc")
    etc_and_sub_dirs = sorted(["/etc", "/etc/pki/tls/private", "/etc/pki/tls/certs",
        "/etc/pki/ovirt-vmconsole", "/etc/nova/migration", "/etc/sysconfig",
        "/etc/cloud/cloud.cfg.d", "/etc/rc.d/init.d"])
    ls_etc = simple_command("/bin/ls -lan {0}".format(' '.join(etc_and_sub_dirs)))
    ls_lib_firmware = simple_command("/bin/ls -lanR /lib/firmware")
    ls_ocp_cni_openshift_sdn = simple_command("/bin/ls -l /var/lib/cni/networks/openshift-sdn")
    ls_origin_local_volumes_pods = simple_command("/bin/ls -l /var/lib/origin/openshift.local.volumes/pods")
    ls_osroot = simple_command("/bin/ls -lan /")
    ls_run_systemd_generator = simple_command("/bin/ls -lan /run/systemd/generator")
    ls_R_var_lib_nova_instances = simple_command("/bin/ls -laR /var/lib/nova/instances")
    ls_sys_firmware = simple_command("/bin/ls -lanR /sys/firmware")
    ls_tmp = simple_command("/bin/ls -la /tmp")
    ls_usr_lib64 = simple_command("/bin/ls -lan /usr/lib64")
    ls_var_lib_mongodb = simple_command("/bin/ls -la /var/lib/mongodb")
    ls_var_lib_nova_instances = simple_command("/bin/ls -laRZ /var/lib/nova/instances")
    ls_var_log = simple_command("/bin/ls -la /var/log /var/log/audit")
    ls_var_opt_mssql = simple_command("/bin/ls -ld /var/opt/mssql")
    ls_var_opt_mssql_log = simple_command("/bin/ls -la /var/opt/mssql/log")
    ls_var_spool_clientmq = simple_command("/bin/ls -ln /var/spool/clientmqueue")
    ls_var_spool_postfix_maildrop = simple_command("/bin/ls -ln /var/spool/postfix/maildrop")
    ls_var_tmp = simple_command("/bin/ls -ln /var/tmp")
    ls_var_run = simple_command("/bin/ls -lnL /var/run")
    ls_var_www = simple_command("/bin/ls -la /dev/null /var/www")  # https://github.com/RedHatInsights/insights-core/issues/827
    lsblk = simple_command("/bin/lsblk")
    lsblk_pairs = simple_command("/bin/lsblk -P -o NAME,KNAME,MAJ:MIN,FSTYPE,MOUNTPOINT,LABEL,UUID,RA,RO,RM,MODEL,SIZE,STATE,OWNER,GROUP,MODE,ALIGNMENT,MIN-IO,OPT-IO,PHY-SEC,LOG-SEC,ROTA,SCHED,RQ-SIZE,TYPE,DISC-ALN,DISC-GRAN,DISC-MAX,DISC-ZERO")
    lscpu = simple_command("/usr/bin/lscpu")
    lsmod = simple_command("/sbin/lsmod")
    lsof = simple_command("/usr/sbin/lsof")
    lspci = simple_command("/sbin/lspci -k")
    lssap = simple_command("/usr/sap/hostctrl/exe/lssap")
    lsscsi = simple_command("/usr/bin/lsscsi")
    lsvmbus = simple_command("/usr/sbin/lsvmbus -vv")
    lvm_conf = simple_file("/etc/lvm/lvm.conf")
    lvs_noheadings = simple_command("/sbin/lvs --nameprefixes --noheadings --separator='|' -a -o lv_name,lv_size,lv_attr,mirror_log,vg_name,devices,region_size,data_percent,metadata_percent,segtype,seg_monitor,lv_kernel_major,lv_kernel_minor --config=\"global{locking_type=0}\"")
    mac_addresses = glob_file("/sys/class/net/*/address")
    machine_id = first_file(["etc/insights-client/machine-id", "etc/redhat-access-insights/machine-id", "etc/redhat_access_proactive/machine-id"])
    mariadb_log = simple_file("/var/log/mariadb/mariadb.log")
    max_uid = simple_command("/bin/awk -F':' '{ if($3 > max) max = $3 } END { print max }' /etc/passwd")

    @datasource()
    def md5chk_file_list(broker):
        """ Provide a list of files to be processed by the ``md5chk_files`` spec """
        return ["/etc/pki/product/69.pem", "/etc/pki/product-default/69.pem", "/usr/lib/libsoftokn3.so", "/usr/lib64/libsoftokn3.so", "/usr/lib/libfreeblpriv3.so", "/usr/lib64/libfreeblpriv3.so"]
    md5chk_files = foreach_execute(md5chk_file_list, "/usr/bin/md5sum %s")
    mdstat = simple_file("/proc/mdstat")
    meminfo = first_file(["/proc/meminfo", "/meminfo"])
    messages = simple_file("/var/log/messages")
    modinfo_i40e = simple_command("/sbin/modinfo i40e")
    modinfo_igb = simple_command("/sbin/modinfo igb")
    modinfo_ixgbe = simple_command("/sbin/modinfo ixgbe")
    modinfo_veth = simple_command("/sbin/modinfo veth")
    modinfo_vmxnet3 = simple_command("/sbin/modinfo vmxnet3")
    modprobe = glob_file(["/etc/modprobe.conf", "/etc/modprobe.d/*.conf"])
    mokutil_sbstate = simple_command("/bin/mokutil --sb-state")
    mongod_conf = glob_file([
                            "/etc/mongod.conf",
                            "/etc/mongodb.conf",
                            "/etc/opt/rh/rh-mongodb26/mongod.conf"
                            ])
    mount = simple_command("/bin/mount")
    mounts = simple_file("/proc/mounts")
    mssql_conf = simple_file("/var/opt/mssql/mssql.conf")
    multicast_querier = simple_command("/usr/bin/find /sys/devices/virtual/net/ -name multicast_querier -print -exec cat {} \;")
    multipath_conf = simple_file("/etc/multipath.conf")
    multipath_conf_initramfs = simple_command("/bin/lsinitrd -f /etc/multipath.conf")
    multipath__v4__ll = simple_command("/sbin/multipath -v4 -ll")
    mysqladmin_vars = simple_command("/bin/mysqladmin variables")
    mysql_log = glob_file([
                          "/var/log/mysql/mysqld.log",
                          "/var/log/mysql.log",
                          "/var/opt/rh/rh-mysql*/log/mysql/mysqld.log"
                          ])
    named_checkconf_p = simple_command("/usr/sbin/named-checkconf -p")
    named_conf = simple_file("/etc/named.conf")
    namespace = simple_command("/bin/ls /var/run/netns")
    ndctl_list_Ni = simple_command("/usr/bin/ndctl list -Ni")
    netconsole = simple_file("/etc/sysconfig/netconsole")
    netstat = simple_command("/bin/netstat -neopa")
    netstat_agn = simple_command("/bin/netstat -agn")
    netstat_i = simple_command("/bin/netstat -i")
    netstat_s = simple_command("/bin/netstat -s")
    networkmanager_conf = simple_file("/etc/NetworkManager/NetworkManager.conf")
    networkmanager_dispatcher_d = glob_file("/etc/NetworkManager/dispatcher.d/*-dhclient")
    neutron_conf = first_file(["/var/lib/config-data/puppet-generated/neutron/etc/neutron/neutron.conf", "/etc/neutron/neutron.conf"])
    neutron_sriov_agent = first_file([
        "/var/lib/config-data/puppet-generated/neutron/etc/neutron/plugins/ml2/sriov_agent.ini",
        "/etc/neutron/plugins/ml2/sriov_agent.ini"])
    neutron_dhcp_agent_ini = first_file(["/var/lib/config-data/puppet-generated/neutron/etc/neutron/dhcp_agent.ini", "/etc/neutron/dhcp_agent.ini"])
    neutron_l3_agent_ini = first_file(["/var/lib/config-data/puppet-generated/neutron/etc/neutron/l3_agent.ini", "/etc/neutron/l3_agent.ini"])
    neutron_l3_agent_log = first_file(["/var/log/containers/neutron/l3-agent.log", "/var/log/neutron/l3-agent.log"])
    neutron_metadata_agent_ini = first_file(["/var/lib/config-data/puppet-generated/neutron/etc/neutron/metadata_agent.ini", "/etc/neutron/metadata_agent.ini"])
    neutron_metadata_agent_log = first_file(["/var/log/containers/neutron/metadata-agent.log", "/var/log/neutron/metadata-agent.log"])
    neutron_ovs_agent_log = first_file(["/var/log/containers/neutron/openvswitch-agent.log", "/var/log/neutron/openvswitch-agent.log"])
    neutron_plugin_ini = first_file(["/var/lib/config-data/puppet-generated/neutron/etc/neutron/plugin.ini", "/etc/neutron/plugin.ini"])
    nfnetlink_queue = simple_file("/proc/net/netfilter/nfnetlink_queue")
    nfs_exports = simple_file("/etc/exports")
    nfs_exports_d = glob_file("/etc/exports.d/*.exports")
    nginx_conf = glob_file([
                           "/etc/nginx/*.conf", "/etc/nginx/conf.d/*", "/etc/nginx/default.d/*",
                           "/opt/rh/nginx*/root/etc/nginx/*.conf", "/opt/rh/nginx*/root/etc/nginx/conf.d/*", "/opt/rh/nginx*/root/etc/nginx/default.d/*",
                           "/etc/opt/rh/rh-nginx*/nginx/*.conf", "/etc/opt/rh/rh-nginx*/nginx/conf.d/*", "/etc/opt/rh/rh-nginx*/nginx/default.d/*"
                           ])
    nmcli_conn_show = simple_command("/usr/bin/nmcli conn show")
    nmcli_dev_show = simple_command("/usr/bin/nmcli dev show")
    nova_api_log = first_file(["/var/log/containers/nova/nova-api.log", "/var/log/nova/nova-api.log"])
    nova_compute_log = first_file(["/var/log/containers/nova/nova-compute.log", "/var/log/nova/nova-compute.log"])
    nova_conf = first_file([
                           "/var/lib/config-data/puppet-generated/nova/etc/nova/nova.conf",
                           "/var/lib/config-data/puppet-generated/nova_libvirt/etc/nova/nova.conf",
                           "/etc/nova/nova.conf"
                           ])
    nova_crontab = simple_command("/usr/bin/crontab -l -u nova")
    nova_uid = simple_command("/usr/bin/id -u nova")
    nscd_conf = simple_file("/etc/nscd.conf")
    nsswitch_conf = simple_file("/etc/nsswitch.conf")
    ntp_conf = simple_file("/etc/ntp.conf")
    ntpq_leap = simple_command("/usr/sbin/ntpq -c 'rv 0 leap'")
    ntptime = simple_command("/usr/sbin/ntptime")
    numa_cpus = glob_file("/sys/devices/system/node/node[0-9]*/cpulist")
    numeric_user_group_name = simple_command("/bin/grep -c '^[[:digit:]]' /etc/passwd /etc/group")
    nvme_core_io_timeout = simple_file("/sys/module/nvme_core/parameters/io_timeout")
    oc_get_clusterrole_with_config = simple_command("/usr/bin/oc get clusterrole --config /etc/origin/master/admin.kubeconfig")
    oc_get_clusterrolebinding_with_config = simple_command("/usr/bin/oc get clusterrolebinding --config /etc/origin/master/admin.kubeconfig")
    odbc_ini = simple_file("/etc/odbc.ini")
    odbcinst_ini = simple_file("/etc/odbcinst.ini")
    open_vm_tools_stat_raw_text_session = simple_command("/usr/bin/vmware-toolbox-cmd stat raw text session")
    openshift_hosts = simple_file("/root/.config/openshift/hosts")
    openshift_router_pid = simple_command("/usr/bin/pgrep -n openshift-route")
    openshift_router_environ = foreach_collect(openshift_router_pid, "/proc/%s/environ")
    openvswitch_other_config = simple_command("/usr/bin/ovs-vsctl -t 5 get Open_vSwitch . other_config")
    os_release = simple_file("etc/os-release")
    ose_master_config = simple_file("/etc/origin/master/master-config.yaml")
    ose_node_config = simple_file("/etc/origin/node/node-config.yaml")
    ovirt_engine_server_log = simple_file("/var/log/ovirt-engine/server.log")
    ovirt_engine_ui_log = simple_file("/var/log/ovirt-engine/ui.log")
    ovs_vsctl_list_bridge = simple_command("/usr/bin/ovs-vsctl list bridge")
    ovs_vsctl_show = simple_command("/usr/bin/ovs-vsctl show")
    pacemaker_log = first_file(["/var/log/pacemaker.log", "/var/log/pacemaker/pacemaker.log"])
    pci_rport_target_disk_paths = simple_command("/usr/bin/find /sys/devices/ -maxdepth 10 -mindepth 9 -name stat -type f")

    @datasource(Services, context=HostContext)
    def pcp_enabled(broker):
        """ bool: Returns True if pmproxy service is on in services """
        if not broker[Services].is_on("pmproxy"):
            raise SkipComponent("pmproxy not enabled")

    pcp_metrics = simple_command("/usr/bin/curl -s http://127.0.0.1:44322/metrics --connect-timeout 5", deps=[pcp_enabled])
    passenger_status = simple_command("/usr/bin/passenger-status")
    password_auth = simple_file("/etc/pam.d/password-auth")
    pcs_quorum_status = simple_command("/usr/sbin/pcs quorum status")
    pcs_status = simple_command("/usr/sbin/pcs status")
    php_ini = first_file(["/etc/opt/rh/php73/php.ini", "/etc/opt/rh/php72/php.ini", "/etc/php.ini"])
    pluginconf_d = glob_file("/etc/yum/pluginconf.d/*.conf")
    postconf_builtin = simple_command("/usr/sbin/postconf -C builtin")
    postconf = simple_command("/usr/sbin/postconf")
    postgresql_conf = first_file([
                                 "/var/lib/pgsql/data/postgresql.conf",
                                 "/opt/rh/postgresql92/root/var/lib/pgsql/data/postgresql.conf",
                                 "database/postgresql.conf"
                                 ])
    postgresql_log = first_of([
                              glob_file("/var/lib/pgsql/data/pg_log/postgresql-*.log"),
                              glob_file("/opt/rh/postgresql92/root/var/lib/pgsql/data/pg_log/postgresql-*.log"),
                              glob_file("/database/postgresql-*.log")
                              ])
    puppetserver_config = simple_file("/etc/sysconfig/puppetserver")
    proc_netstat = simple_file("proc/net/netstat")
    proc_slabinfo = simple_file("proc/slabinfo")
    proc_snmp_ipv4 = simple_file("proc/net/snmp")
    proc_snmp_ipv6 = simple_file("proc/net/snmp6")
    proc_stat = simple_file("proc/stat")
    pulp_worker_defaults = simple_file("etc/default/pulp_workers")
    pvs_noheadings = simple_command("/sbin/pvs --nameprefixes --noheadings --separator='|' -a -o pv_all,vg_name --config=\"global{locking_type=0}\"")
    qemu_conf = simple_file("/etc/libvirt/qemu.conf")
    qemu_xml = glob_file(r"/etc/libvirt/qemu/*.xml")
    qpid_stat_g = simple_command("/usr/bin/qpid-stat -g --ssl-certificate=/etc/pki/katello/qpid_client_striped.crt -b amqps://localhost:5671")
    qpidd_conf = simple_file("/etc/qpid/qpidd.conf")
    rabbitmq_env = simple_file("/etc/rabbitmq/rabbitmq-env.conf")
    rabbitmq_report = simple_command("/usr/sbin/rabbitmqctl report")
    rabbitmq_startup_log = simple_file("/var/log/rabbitmq/startup_log")
    rabbitmq_users = simple_command("/usr/sbin/rabbitmqctl list_users")
    rc_local = simple_file("/etc/rc.d/rc.local")
    rdma_conf = simple_file("/etc/rdma/rdma.conf")
    readlink_e_etc_mtab = simple_command("/usr/bin/readlink -e /etc/mtab")
    readlink_e_shift_cert_client = simple_command("/usr/bin/readlink -e /etc/origin/node/certificates/kubelet-client-current.pem")
    readlink_e_shift_cert_server = simple_command("/usr/bin/readlink -e /etc/origin/node/certificates/kubelet-server-current.pem")
    redhat_release = simple_file("/etc/redhat-release")
    resolv_conf = simple_file("/etc/resolv.conf")
    rhn_conf = first_file(["/etc/rhn/rhn.conf", "/conf/rhn/rhn/rhn.conf"])
    rhn_entitlement_cert_xml = first_of([glob_file("/etc/sysconfig/rhn/rhn-entitlement-cert.xml*"),
                                   glob_file("/conf/rhn/sysconfig/rhn/rhn-entitlement-cert.xml*")])
    rhn_schema_version = simple_command("/usr/bin/rhn-schema-version")
    rhn_taskomatic_daemon_log = first_file(["/var/log/rhn/rhn_taskomatic_daemon.log",
                                               "rhn-logs/rhn/rhn_taskomatic_daemon.log"])
    rhsm_conf = simple_file("/etc/rhsm/rhsm.conf")
    rhsm_log = simple_file("/var/log/rhsm/rhsm.log")
    rhsm_releasever = simple_file('/var/lib/rhsm/cache/releasever.json')
    rndc_status = simple_command("/usr/sbin/rndc status")
    rpm_V_packages = simple_command("/bin/rpm -V coreutils procps procps-ng shadow-utils passwd sudo chrony", keep_rc=True)
    rsyslog_conf = glob_file(["/etc/rsyslog.conf", "/etc/rsyslog.d/*.conf"])
    samba = simple_file("/etc/samba/smb.conf")

    @datasource(Sap)
    def sap_sid(broker):
        sap = broker[Sap]
        return [sap.sid(i).lower() for i in sap.local_instances]

    sap_hdb_version = foreach_execute(sap_sid, "/usr/bin/sudo -iu %sadm HDB version", keep_rc=True)
    saphostctl_getcimobject_sapinstance = simple_command("/usr/sap/hostctrl/exe/saphostctrl -function GetCIMObject -enuminstances SAPInstance")
    sat5_insights_properties = simple_file("/etc/redhat-access/redhat-access-insights.properties")
    satellite_mongodb_storage_engine = simple_command("/usr/bin/mongo pulp_database --eval 'db.serverStatus().storageEngine'")
    satellite_version_rb = simple_file("/usr/share/foreman/lib/satellite/version.rb")
    satellite_custom_hiera = simple_file("/etc/foreman-installer/custom-hiera.yaml")
    scsi = simple_file("/proc/scsi/scsi")
    scsi_eh_deadline = glob_file('/sys/class/scsi_host/host[0-9]*/eh_deadline')
    scsi_fwver = glob_file('/sys/class/scsi_host/host[0-9]*/fwrev')
    scsi_mod_use_blk_mq = simple_file("/sys/module/scsi_mod/parameters/use_blk_mq")
    sctp_asc = simple_file('/proc/net/sctp/assocs')
    sctp_eps = simple_file('/proc/net/sctp/eps')
    sctp_snmp = simple_file('/proc/net/sctp/snmp')
    sealert = simple_command('/usr/bin/sealert -l "*"')
    selinux_config = simple_file("/etc/selinux/config")
    sestatus = simple_command("/usr/sbin/sestatus -b")
    setup_named_chroot = simple_file("/usr/libexec/setup-named-chroot.sh")
    smbstatus_p = simple_command("/usr/bin/smbstatus -p")
    smartpdc_settings = simple_file("/etc/smart_proxy_dynflow_core/settings.yml")
    sockstat = simple_file("/proc/net/sockstat")
    softnet_stat = simple_file("proc/net/softnet_stat")
    software_collections_list = simple_command('/usr/bin/scl --list')
    spamassassin_channels = simple_command("/bin/grep -r '^\\s*CHANNELURL=' /etc/mail/spamassassin/channel.d")
    ss = simple_command("/usr/sbin/ss -tupna")
    ssh_config = simple_file("/etc/ssh/ssh_config")
    ssh_config_d = glob_file(r"/etc/ssh/ssh_config.d/*.conf")
    ssh_foreman_proxy_config = simple_file("/usr/share/foreman-proxy/.ssh/ssh_config")
    sshd_config = simple_file("/etc/ssh/sshd_config")
    sshd_config_perms = simple_command("/bin/ls -l /etc/ssh/sshd_config")
    sssd_config = simple_file("/etc/sssd/sssd.conf")
    subscription_manager_id = simple_command("/usr/sbin/subscription-manager identity")  # use "/usr/sbin" here, BZ#1690529
    swift_object_expirer_conf = first_file(["/var/lib/config-data/puppet-generated/swift/etc/swift/object-expirer.conf", "/etc/swift/object-expirer.conf"])
    swift_proxy_server_conf = first_file(["/var/lib/config-data/puppet-generated/swift/etc/swift/proxy-server.conf", "/etc/swift/proxy-server.conf"])
    sysconfig_kdump = simple_file("etc/sysconfig/kdump")
    sysconfig_libvirt_guests = simple_file("etc/sysconfig/libvirt-guests")
    sysconfig_network = simple_file("etc/sysconfig/network")
    sysconfig_ntpd = simple_file("/etc/sysconfig/ntpd")
    sysconfig_prelink = simple_file("/etc/sysconfig/prelink")
    sysconfig_sshd = simple_file("/etc/sysconfig/sshd")
    sysconfig_virt_who = simple_file("/etc/sysconfig/virt-who")
    sysctl = simple_command("/sbin/sysctl -a")
    sysctl_conf = simple_file("/etc/sysctl.conf")
    systemctl_cat_rpcbind_socket = simple_command("/bin/systemctl cat rpcbind.socket")
    systemctl_cinder_volume = simple_command("/bin/systemctl show openstack-cinder-volume")
    systemctl_httpd = simple_command("/bin/systemctl show httpd")
    systemctl_nginx = simple_command("/bin/systemctl show nginx")
    systemctl_list_unit_files = simple_command("/bin/systemctl list-unit-files")
    systemctl_list_units = simple_command("/bin/systemctl list-units")
    systemctl_mariadb = simple_command("/bin/systemctl show mariadb")
    systemctl_qpidd = simple_command("/bin/systemctl show qpidd")
    systemctl_qdrouterd = simple_command("/bin/systemctl show qdrouterd")
    systemctl_show_all_services = simple_command("/bin/systemctl show *.service")
    systemctl_show_target = simple_command("/bin/systemctl show *.target")
    systemctl_smartpdc = simple_command("/bin/systemctl show smart_proxy_dynflow_core")
    systemd_analyze_blame = simple_command("/bin/systemd-analyze blame")
    systemd_docker = simple_command("/usr/bin/systemctl cat docker.service")
    systemd_logind_conf = simple_file("/etc/systemd/logind.conf")
    systemd_openshift_node = simple_command("/usr/bin/systemctl cat atomic-openshift-node.service")
    systemd_system_conf = simple_file("/etc/systemd/system.conf")
    systemid = first_of([
        simple_file("/etc/sysconfig/rhn/systemid"),
        simple_file("/conf/rhn/sysconfig/rhn/systemid")
    ])
    systool_b_scsi_v = simple_command("/bin/systool -b scsi -v")
    testparm_s = simple_command("/usr/bin/testparm -s")
    testparm_v_s = simple_command("/usr/bin/testparm -v -s")
    tags = simple_file("/tags.json", kind=RawFileProvider)
    thp_use_zero_page = simple_file("/sys/kernel/mm/transparent_hugepage/use_zero_page")
    thp_enabled = simple_file("/sys/kernel/mm/transparent_hugepage/enabled")
    tmpfilesd = glob_file(["/etc/tmpfiles.d/*.conf", "/usr/lib/tmpfiles.d/*.conf", "/run/tmpfiles.d/*.conf"])
    tomcat_vdc_fallback = simple_command("/usr/bin/find /usr/share -maxdepth 1 -name 'tomcat*' -exec /bin/grep -R -s 'VirtualDirContext' --include '*.xml' '{}' +")
    tuned_adm = simple_command("/usr/sbin/tuned-adm list")
    udev_fc_wwpn_id_rules = simple_file("/usr/lib/udev/rules.d/59-fc-wwpn-id.rules")
    uname = simple_command("/usr/bin/uname -a")
    up2date = simple_file("/etc/sysconfig/rhn/up2date")
    up2date_log = simple_file("/var/log/up2date")
    uptime = simple_command("/usr/bin/uptime")
    usr_journald_conf_d = glob_file(r"usr/lib/systemd/journald.conf.d/*.conf")  # note that etc_journald.conf.d also exists
    vdo_status = simple_command("/usr/bin/vdo status")
    vgdisplay = simple_command("/sbin/vgdisplay")
    vhost_net_zero_copy_tx = simple_file("/sys/module/vhost_net/parameters/experimental_zcopytx")
    vdsm_log = simple_file("var/log/vdsm/vdsm.log")
    vdsm_logger_conf = simple_file("etc/vdsm/logger.conf")
    vma_ra_enabled = simple_file("/sys/kernel/mm/swap/vma_ra_enabled")
    vgs_noheadings = simple_command("/sbin/vgs --nameprefixes --noheadings --separator='|' -a -o vg_all --config=\"global{locking_type=0}\"")
    virsh_list_all = simple_command("/usr/bin/virsh --readonly list --all")
    virt_what = simple_command("/usr/sbin/virt-what")
    virt_who_conf = glob_file([r"etc/virt-who.conf", r"etc/virt-who.d/*.conf"])
    virtlogd_conf = simple_file("/etc/libvirt/virtlogd.conf")
    vsftpd = simple_file("/etc/pam.d/vsftpd")
    vsftpd_conf = simple_file("/etc/vsftpd/vsftpd.conf")
    x86_pti_enabled = simple_file("sys/kernel/debug/x86/pti_enabled")
    x86_ibpb_enabled = simple_file("sys/kernel/debug/x86/ibpb_enabled")
    x86_ibrs_enabled = simple_file("sys/kernel/debug/x86/ibrs_enabled")
    x86_retp_enabled = simple_file("sys/kernel/debug/x86/retp_enabled")
    xinetd_conf = glob_file(["/etc/xinetd.conf", "/etc/xinetd.d/*"])
    yum_conf = simple_file("/etc/yum.conf")
    yum_list_available = simple_command("yum -C --noplugins list available")
    yum_log = simple_file("/var/log/yum.log")
    yum_repolist = simple_command("/usr/bin/yum -C --noplugins repolist")
    yum_repos_d = glob_file("/etc/yum.repos.d/*.repo")
    zipl_conf = simple_file("/etc/zipl.conf")
    rpm_format = format_rpm()
    installed_rpms = simple_command("/bin/rpm -qa --qf '%s'" % rpm_format, context=HostContext)



#!/usr/bin/env python
import itertools
import os
from collections import defaultdict

import pandas as pd

from ansible.parsing.dataloader import DataLoader
from ansible.inventory.manager import InventoryManager

from insights.core import dr, plugins
from insights.core.archives import extract
from insights.core.hydration import create_context
from insights.specs import Specs


ID_GENERATOR = itertools.count()


class ClusterMeta(dict):
    def __init__(self, num_members, kwargs):
        self.num_members = num_members
        self.update(**kwargs)


@plugins.combiner(optional=[Specs.machine_id, Specs.hostname])
def machine_id(mid, hn):
    ds = mid or hn
    if ds:
        return ds.content[0].strip()
    return str(next(ID_GENERATOR))


def parse_inventory(path):
    inventory = InventoryManager(loader=DataLoader(), sources=path)
    return inventory.get_groups_dict()


def attach_machine_id(result, mid):
    key = "machine_id"
    if isinstance(result, list):
        for r in result:
            r[key] = mid
    else:
        result[key] = mid
    return result


def process_archives(graph, archives):
    for archive in archives:
        if os.path.isfile(archive):
            with extract(archive) as ex:
                ctx = create_context(ex.tmp_dir)
                broker = dr.Broker()
                broker[ctx.__class__] = ctx
                yield dr.run(broker=broker)
        else:
            ctx = create_context(archive)
            broker = dr.Broker()
            broker[ctx.__class__] = ctx
            yield dr.run(graph, broker=broker)


def extract_facts(brokers):
    results = defaultdict(list)
    for b in brokers:
        mid = b[machine_id]
        for k, v in b.get_by_type(plugins.fact).items():
            r = attach_machine_id(v, mid)
            if isinstance(r, list):
                results[k].extend(r)
            else:
                results[k].append(r)
    return results


def process_facts(facts, meta, broker, cluster_graph):
    broker[ClusterMeta] = meta
    for k, v in facts.items():
        broker[k] = pd.DataFrame(v)
    return dr.run(cluster_graph, broker=broker)


def process_cluster(graph, archives, broker, inventory=None):
    host_graph = dict((k, v) for k, v in graph.items() if k in dr.COMPONENTS[dr.GROUPS.single])
    host_graph[machine_id] = dr.DELEGATES[machine_id].dependencies
    cluster_graph = dict((k, v) for k, v in graph.items() if k not in host_graph)

    inventory = parse_inventory(inventory) if inventory else {}

    brokers = process_archives(host_graph, archives)
    facts = extract_facts(brokers)
    meta = ClusterMeta(len(archives), inventory)

    return process_facts(facts, meta, broker, cluster_graph)



from __future__ import print_function
from insights.parsers import lvm
from insights.tests import context_wrap
from .lvm_test_data import LVMCONFIG

WARNINGS_CONTENT = """
WARNING
valid data 1
Checksum Error
valid data 2
  Failed to write
  Attempt To Close Device
valid data 3
""".strip()

WARNINGS_FOUND = """
WARNING
Checksum Error
  Failed to write
  Attempt To Close Device
""".strip()


def test_find_warnings():
    data = [l for l in lvm.find_warnings(WARNINGS_CONTENT.splitlines())]
    assert len(data) == len(WARNINGS_FOUND.splitlines())
    assert data == WARNINGS_FOUND.splitlines()


def compare_partial_dicts(result, expected):
    """
    Make sure all the keys in expected are matched by keys in result, and
    that the values stored in those keys match.  Result can contain more
    items than expected - those are ignored.

    Used in the test_lvs, test_pvs and test_vgs tests.
    """
    # return all(result[k] == expected[k] for k in expected.keys())
    mismatches = 0
    for k in expected.keys():
        if not result[k] == expected[k]:
            print("Failed for key {k}, {r} != {e}".format(k=k, r=result[k], e=expected[k]))
            mismatches += 1
    return mismatches == 0


def test_lvmconfig():
    p = lvm.LvmConfig(context_wrap(LVMCONFIG))
    assert p.data["dmeventd"]["raid_library"] == "libdevmapper-event-lvm2raid.so"
    assert p.data["global"]["thin_check_options"] == ["-q", "--clear-needs-check-flag"]



import pytest
from ...tests import context_wrap
from ..chkconfig import ChkConfig

SERVICES = """
auditd         	0:off	1:off	2:on	3:on	4:on	5:on	6:off
crond          	0:off	1:off	2:on	3:on	4:on	5:on	6:off
iptables       	0:off	1:off	2:on	3:on	4:on	5:on	6:off
kdump          	0:off	1:off	2:off	3:on	4:on	5:on	6:off
restorecond    	0:off	1:off	2:off	3:off	4:off	5:off	6:off
xinetd          0:off   1:off   2:on    3:on    4:on    5:on    6:off
        rexec:          off
        rlogin:         off
        rsh:            on
        tcpmux-server:  off
        telnet:         on
""".strip()

RHEL_73_SERVICES = """
Note: This output shows SysV services only and does not include native
      systemd services. SysV configuration data might be overridden by native
      systemd configuration.

      If you want to list systemd services use 'systemctl list-unit-files'.
      To see services enabled on particular target use
      'systemctl list-dependencies [target]'.

netconsole     	0:off	1:off	2:off	3:off	4:off	5:off	6:off
network        	0:off	1:off	2:on	3:on	4:on	5:on	6:off
rhnsd          	0:off	1:off	2:on	3:on	4:on	5:on	6:off

xinetd based services:
        chargen-dgram:  off
        chargen-stream: off
        daytime-dgram:  off
        daytime-stream: off
        discard-dgram:  off
        discard-stream: off
        echo-dgram:     off
        echo-stream:    off
        rsync:          on
        tcpmux-server:  off
        time-dgram:     off
        time-stream:    off
"""


def test_chkconfig():
    context = context_wrap(SERVICES)
    chkconfig = ChkConfig(context)
    assert len(chkconfig.services) == 11
    assert len(chkconfig.parsed_lines) == 11
    assert chkconfig.is_on('crond')
    assert chkconfig.is_on('kdump')
    assert chkconfig.is_on('telnet')
    assert not chkconfig.is_on('restorecond')
    assert not chkconfig.is_on('rlogin')


def test_levels_on():
    chkconfig = ChkConfig(context_wrap(SERVICES))
    assert chkconfig.levels_on('crond') == set(['2', '3', '4', '5'])
    assert chkconfig.levels_on('telnet') == set(['2', '3', '4', '5'])
    assert chkconfig.levels_on('rlogin') == set([])
    with pytest.raises(KeyError):
        assert chkconfig.levels_on('bad_name')


def test_levels_off():
    chkconfig = ChkConfig(context_wrap(SERVICES))
    assert chkconfig.levels_off('crond') == set(['0', '1', '6'])
    assert chkconfig.levels_off('telnet') == set(['0', '1', '6'])
    assert chkconfig.levels_off('rlogin') == set(['0', '1', '2', '3',
                                                    '4', '5', '6'])
    with pytest.raises(KeyError):
        assert chkconfig.levels_off('bad_name')


def test_rhel_73():
    chkconfig = ChkConfig(context_wrap(RHEL_73_SERVICES))
    assert sorted(chkconfig.level_states.keys()) == sorted([
        'netconsole', 'network', 'rhnsd', 'chargen-dgram', 'chargen-stream',
        'daytime-dgram', 'daytime-stream', 'discard-dgram', 'discard-stream',
        'echo-dgram', 'echo-stream', 'rsync', 'tcpmux-server', 'time-dgram',
        'time-stream',
    ])
    assert chkconfig.levels_off('netconsole') == set(['0', '1', '2', '3', '4', '5', '6'])
    assert chkconfig.levels_on('network') == set(['2', '3', '4', '5'])
    assert chkconfig.levels_on('rsync') == set(['0', '1', '2', '3', '4', '5', '6'])



"""
Parsers for ``ip`` command outputs
==================================

This module provides the following parsers:

IpAddr - command ``ip addr``
----------------------------

RouteDevices - command ``ip route show table all``
--------------------------------------------------

IpNeighParser - command ``ip neigh show nud all``
-------------------------------------------------

IpLinkInfo - command ``ip -d -s link``
--------------------------------------
"""

from __future__ import print_function

import six
from collections import defaultdict, deque
from .. import parser, CommandParser
from ..contrib import ipaddress
from insights.specs import Specs


class NetworkInterface(object):
    def __init__(self, d):
        self.data = d
        addresses = [u"/".join([a["addr"], a["mask"]]) for a in self.data["addr"]]
        self.addresses = list(map(ipaddress.ip_interface, addresses))

    def __len__(self):
        return len(self.addresses)

    def __lt__(self, other):
        return self["name"] < other["name"]

    def __eq__(self, other):
        return self["name"] == other["name"]

    def __getitem__(self, item):
        return self.data[item]

    def addrs(self, version=None):
        if version:
            return [str(a.ip) for a in self.addresses if a.version == version]
        else:
            return [str(a.ip) for a in self.addresses]


def parse_ip_addr(content):
    r = {}
    current = {}
    if_details = {}
    rx_next_line = False
    tx_next_line = False
    content = [l.strip() for l in content if "Message truncated" not in l]
    for line in filter(None, content):
        if rx_next_line and current:
            parse_rx_stats(line, current)
            rx_next_line = False
        if tx_next_line and current:
            parse_tx_stats(line, current)
            tx_next_line = False
        elif line[0].isdigit() and "state" in line:
            current = parse_interface(line)
            r[current["name"]] = current
        elif line.startswith("link"):
            parse_link(line, current)
        elif 'vxlan' in line:
            split_content = line.split()
            current['vxlan'] = split_content
        elif 'openvswitch' in line:
            split_content = line.split()
            current['openvswitch'] = split_content
        elif 'geneve' in line:
            split_content = line.split()
            current['geneve'] = split_content
        elif line.startswith("inet"):
            parse_inet(line, current)
        elif line.startswith("RX"):
            rx_next_line = True
        elif line.startswith("TX"):
            tx_next_line = True
    for k, v in r.items():
        if_details[k] = NetworkInterface(v)
    return if_details


def parse_interface(line):
    split_content = line.split()
    idx, name, _ = line.split(":", 2)
    virtual = "@" in name
    if virtual:
        name, physical_name = name.split("@")
    current = {
        "index": int(idx),
        "name": name.strip(),
        "physical_name": physical_name if virtual else None,
        "virtual": virtual,
        "flags": split_content[2].strip("<>").split(","),
        "addr": []
    }
    # extract properties
    for i in range(3, len(split_content), 2):
        key, value = (split_content[i], split_content[i + 1])
        current[key] = int(value) if key in ["mtu", "qlen"] else value
    return current


def parse_link(line, d):
    split_content = line.split()
    d["type"] = split_content[0].split("/")[1]
    if "peer" in line and len(split_content) >= 3:
        d["peer_ip"] = split_content[1]
        d["peer"] = split_content[3]
    elif len(split_content) >= 2:
        d["mac"] = split_content[1]


def parse_inet(line, d):
    split_content = line.split()
    p2p = "peer" in split_content
    addr, mask = split_content[3 if p2p else 1].split("/")
    d["addr"].append({
        "addr": addr,
        "mask": mask,
        "local_addr": split_content[1] if p2p else None,
        "p2p": p2p
    })


def parse_rx_stats(line, d):
    split_content = line.split()
    d["rx_bytes"] = int(split_content[0])
    d["rx_packets"] = int(split_content[1])
    d["rx_errors"] = int(split_content[2])
    d["rx_dropped"] = int(split_content[3])
    d["rx_overrun"] = int(split_content[4])
    d["rx_mcast"] = int(split_content[5])


def parse_tx_stats(line, d):
    split_content = line.split()
    d["tx_bytes"] = int(split_content[0])
    d["tx_packets"] = int(split_content[1])
    d["tx_errors"] = int(split_content[2])
    d["tx_dropped"] = int(split_content[3])
    d["tx_carrier"] = int(split_content[4])
    d["tx_collsns"] = int(split_content[5])


@parser(Specs.ip_addr)
class IpAddr(CommandParser):
    """
    This parser reads the output of ``ip addr`` into a dict whose key is
    the interface name.  The information about this interface`addr` key is a array to store all address.
    Different type have different output. Peer ip and general interface have difference type.

    Example output::

        1: lo: <LOOPBACK,UP,LOWER_UP> mtu 16436 qdisc noqueue state UNKNOWN
            link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00
            inet 127.0.0.1/8 scope host lo
            inet6 ::1/128 scope host
            valid_lft forever preferred_lft forever

    Resultant data structure::

        {
            "index": 1,
            "physical_name": null,
            "qdisc": "noqueue",
            "name": "lo",
            "state": "UNKNOWN",
            "virtual": false,
            "mtu": 16436,
            "mac": "00:00:00:00:00:00",
            "flags": [
                "LOOPBACK",
                "UP",
                "LOWER_UP"
            ],
            "type": "loopback",
            "addr": [
                {
                    "local_addr": null,
                    "mask": "8",
                    "p2p": false,
                    "addr": "127.0.0.1"
                },
                {
                    "local_addr": null,
                    "mask": "128",
                    "p2p": false,
                    "addr": "::1"
                }
            ]
        }

    Examples:

        >>> for iface in shared[IpAddr]:
        ...     print 'Interface:', iface['name']
        ...     print 'State:', iface['state']
        ...     print 'Addresses:', ', '.join(a['addr'] for a in iface['addr'])
        ...
        Interface: lo
        State: UNKNOWN
        Addresses: 127.0.0.1, ::1
        Interface: eth7
        State: DOWN
        Addresses:
        Interface: tunl1
        State: DOWN
        Addresses: 172.30.0.1
        Interface: bond1.57
        State: UP
        Addresses:10.192.4.171, fe80::211:3fff:fee2:f59e, 2001::211:3fff:fee2:f59e
        Interface: ip.tun2
        State: UNKNOWN
        Addresses: 192.168.122.6

    """

    def parse_content(self, content):
        self.data = parse_ip_addr(content)

    def __iter__(self):
        """
        (iterable): Iterate through the list of available interfaces (in no order)
        """
        return iter(self.data.values())

    def __len__(self):
        """
        (int): number of interfaces configured.
        """
        return len(self.data)

    def __getitem__(self, item):
        """
        Parameters:
            item (str): Interface name
        Returns:
            (dict): Dictionary of the named interface
        """
        return self.data[item]

    def __contains__(self, item):
        """
        Is the given interface name configured?
        """
        return item in self.data

    @property
    def active(self):
        """
        (list): List of interfaces with 'UP' set in their flags.
        """
        return [i["name"] for i in self if "UP" in i["flags"]]


class Route(object):
    def __init__(self, data):
        for k, v in data.items():
            setattr(self, k, v)

    def __repr__(self):
        return self.__dict__.__repr__()


@parser(Specs.ip_route_show_table_all)
class RouteDevices(CommandParser):
    """
    This parser reads the output of the command ``ip route show table all``
    and provides access to the routing table.

    Each object in these properties is stored as a Route object, which has
    as properties the various parts of the route line.  Route objects will
    always include these properties (all defaulting to None):

    * ``type`` - the type of route (e.g. 'throw')
    * ``dev`` - the device routed through (e.g. 'eth0')
    * ``via`` - the external address routed through, or None if direct
    * ``netmask`` - the CIDR notation of the network as a string (e.g. '24')
    * ``table`` - the routing table, or None if not given.

    Destinations given as **broadcast**, **throw**, **unreachable**,
    **prohibit**, **blackhole** and **nat** are discarded.

    Destinations given as **unicast**, **multicast** and **local** are considered.

    Properties::

        by_device (dict): routes by device (e.g. 'eth0').
        by_type (dict): routes by table type (e.g. 'throw', default = 'None').
        by_table (dict): routes by table (e.g. 'mgmt', default = 'None').

    Sample routing table::

        default via 192.168.1.254 dev enp0s25
        10.0.0.0/8 via 10.64.54.1 dev tun0  proto static  metric 50
        10.64.54.0/23 dev tun0  proto kernel  scope link  src 10.64.54.44  metric 50
        66.187.239.220 via 192.168.23.250 dev enp0s25  proto static
        192.168.0.0/24 dev enp0s25  proto kernel  scope link  src 192.168.1.37
        192.168.122.0/24 dev virbr0  proto kernel  scope link  src 192.168.122.1

    Examples:

        >>> routes = shared[RouteDevices]
        >>> 'default' in routes
        True
        >>> '10.64.54.0/23' in routes
        False
        >>> len(routes['10.64.54.0/23']) # Multiple routes possible, esp. for default
        1
        >>> tunnel = routes['10.64.54.0/23'][0]
        >>> tunnel.dev
        'tun0'
        >>> tunnel.proto
        'kernel'
        >>> len(routes.by_device['tun0'])
        2
        >>> routes.by_device['tun0'][1] == tunnel
        True

    """
    SAVED_TYPES = set(["unicast", "multicast", 'local'])

    # Why have types we ignore?
    IGNORE_TYPES = set(["broadcast",
                        "throw",
                        "unreachable",
                        "prohibit",
                        "blackhole",
                        "nat"])

    @property
    def by_prefix(self):
        """
        (Route): The dictionary of routes by prefix  (e.g. '192.168.0.0/24').
        """
        return self.routes.get('by_prefix', {})

    @property
    def defaults(self):
        """
        (list): The list of default routes.
        """
        return self.routes.get('by_prefix', {}).get('default', [])

    def __contains__(self, prefix):
        """
        (bool): Is there a route with the given prefix?
        """
        return prefix in self.data

    def __getitem__(self, prefix):
        """
        (Route): Retrieve the given prefix - same as the ``by_prefix`` dict.
        """
        return self.data.get(prefix, None)

    def parse_content(self, content):
        """
        Read the routing table and construct the routes and data properties.
        """
        self.data = defaultdict(list)
        prev_line = ''
        for line in content:
            # Seems to not get blank lines here...
            # Leading spaces indicate a line continued from the previous
            if line.startswith('    '):
                prev_line += line
            else:
                if prev_line:
                    self.parse_line(prev_line)
                prev_line = line
        # Grab last line
        self.parse_line(prev_line)

        # Reprocess these into routes by prefix, device, type and table.
        self.routes = defaultdict(lambda: defaultdict(list))
        all_routes = [r for routes in self.data.values() for r in routes]
        # For some reason, if we try to construct the by_prefix property
        # as a defaultdict(list) here we break the tests.  Leaving it as is...
        self.routes['by_prefix'] = self.data
        self.by_device = defaultdict(list)
        self.by_type = defaultdict(list)
        self.by_table = defaultdict(list)
        for route in all_routes:
            table_type = route.type if route.type else 'None'
            dev = route.dev if route.dev else 'None'
            table = route.table if route.table else 'None'
            self.by_device[dev].append(route)
            self.by_type[table_type].append(route)
            self.by_table[table].append(route)
        # self.routes = {k: dict(v) for k, v in self.routes.items()}

    def parse_line(self, line):
        parts = deque(line.split(None))
        route = self.parse_route(parts)
        if route and (route.prefix != 'default' or not route.table):
            self.data[route.prefix].append(route)

    def parse_route(self, parts):
        required_parts = ['via', 'dev', 'type', 'netmask', 'prefix', 'table']
        route = dict((part, None) for part in required_parts)
        table_type = None
        if parts[0] in self.IGNORE_TYPES:
            return None
        if parts[0] in self.SAVED_TYPES:
            table_type = parts.popleft()
        route['type'] = table_type
        prefix = parts.popleft()
        route['netmask'] = 255
        if '/' in prefix:
            route['netmask'] = int(prefix.split('/')[1])
        route['prefix'] = prefix
        self.parse_info_spec(parts, route)
        self.parse_node_spec(parts, route)
        return Route(route)

    def parse_info_spec(self, parts, route):
        keys = ['via', 'dev']
        for k in keys:
            route[k] = None
        if not parts:
            return
        for key in keys:
            if len(parts) > 1 and parts[0] == key:
                k, v = parts.popleft(), parts.popleft()
                route[k] = v

    def parse_node_spec(self, parts, route):
        while parts:
            if parts[0] == 'cache':
                route['cache'] = True
                parts.popleft()
                continue
            if len(parts) == 1:
                return
            k, v = parts.popleft(), parts.popleft()
            route[k] = v

    def ifaces(self, ip):
        """
        Given an IP address, choose the best iface name to return.  If
        there are multiple routes that match, then the one with the most
        specific netmask will be returned.  There may be multiple interfaces
        that serve this route so it returns a list.  If there are default
        routes, then these are used if a route is not found.  If no default
        routes are found, then return ``None``.

        Returns:
            (list): Device names that serve this network, or None if not found.

        Examples:

            >>> ip_table = shared[RouteDevices]
            >>> iface = ip_table.ifaces(YOUR_IP_ADDRESS_STRING)
        """
        if ip is None:
            return
        routes = self.by_type.get('None', [])
        addr = ipaddress.ip_address(six.u(ip))
        # Iterate through by descending netmask, so first found is most precise
        for route in sorted(routes, key=lambda r: r.netmask, reverse=True):
            if route.prefix == "default":
                continue
            net = ipaddress.ip_network(six.u(route.prefix))
            # Only test containment if this is the same verison of IP address.
            if addr.version != net.version:
                continue
            if addr not in net:
                continue
            return [r.dev for r in self.by_prefix[route.prefix] if r.dev]

        if self.defaults:
            return [self.defaults[0].dev]

        return None


class IpNeighParser(CommandParser):
    """
    This parser takes the output of ``ip neigh show nud all`` results for
    ARP and NDISC cache entries and reads them into a dictionary of results
    keyed on the name of the IP address.  Each item is a dictionary of the
    properties of that address record in the ARP table.  Fields usually
    include:

    * ``dev`` - the device the address was seen on
    * ``lladdr`` - the link level (MAC) address associated with the IP
    * ``nud`` - the Neighbour Unreachability Detection result, which is one
      of the following values:

      * **permanent** - the neighbour is permanently valid
      * **noarp** - the neighbour is valid and does not need revalidation
      * **reachable** - the neighbour is valid until its lifetime expires
      * **stale** - the neighour has not been seen in a while and its
        lifetime has expired, but it may still be valid
      * **failed** - the neighbour resolution has failed

    This class deals with both IPv4 and IPv6 records, and is subclassed to
    the two related parsers.

    Sample (IPv4) input data::

        172.17.0.19 dev docker0  FAILED
        172.17.42.1 dev lo lladdr 00:00:00:00:00:00 NOARP

    Examples:

        >>> neighb4 = shared[Ipv4Neigh]
        >>> '192.168.0.1' in neighb4
        False
        >>> '172.17.0.19' in neighb4
        True
        >>> neighb4['172.17.0.19']['dev']
        'docker0'
        >>> neighb4['172.17.42.1']['lladdr']
        '00:00:00:00:00:00'
    """

    VALID_NUD_STATES = {
        'PERMANENT': 0,
        'NOARP': 1,
        'REACHABLE': 2,
        'STALE': 3,
        'DELAY': 4,
        'FAILED': 5,
        'INCOMPLETE': 6,
    }

    def parse_content(self, content):
        """
        Parse the lines if the ``ip neighbor`` output.  Each line is split up
        into words on spaces, and must meet the following criteria:

        * there must be at least two words on the line
        * the first word must be a valid IP (v4 or v6) address
        * the last word must be a valid Neighbour Unreachability Detection state
        * the remaining words must be in pairs - these are then combined into
          key-value pairs for a dictionary.

        Each line is then stored in a dictionary by the address (as a string).
        The address as parsed by the ``ipaddress`` module is stored in the
        ``addr`` item in the dictionary, for convenience.
        """
        self.data = {}
        self.unparsed_lines = []
        for line in filter(None, content):
            split_result = line.split()
            # Need at least IP address, something, and reachability
            if len(split_result) < 2:
                self.unparsed_lines.append({
                    'line': line,
                    'reason': "not enough words"
                })
                continue
            # Total words needs to be even: beginning + 2*keyvals + ending
            if len(split_result) % 2 == 1:
                self.unparsed_lines.append({
                    'line': line,
                    'reason': "odd number of words"
                })
                continue
            # Don't parse this line if the first thing isn't an
            # IP address
            try:
                addr = ipaddress.ip_address(six.u(split_result[0]))
            except ValueError:
                self.unparsed_lines.append({
                    'line': line,
                    'reason': "can't convert address '" + split_result[0] + "'"
                })
                continue
            # Don't parse this line if the last item doesn't seem to be a
            # neighbour unreachability state
            if split_result[-1] not in self.VALID_NUD_STATES:
                self.unparsed_lines.append({
                    'line': line,
                    'reason': split_result[-1] + " is not a valid state"
                })
                continue

            # OK, good to go, split everything in the middle up
            key_value_content = split_result[1:-1]
            if len(key_value_content) >= 2:
                entry = dict((k, v) for k, v in zip(key_value_content[0::2],
                                                    key_value_content[1::2]))
            else:
                entry = {}
            entry["nud"] = split_result[-1]
            entry['addr'] = addr  # save the object
            self.data[split_result[0]] = entry

    def __contains__(self, item):
        """
        (bool): Is there neighbour information for the given address?
        """
        return item in self.data

    def __getitem__(self, item):
        """
        (dict): Get the neighbour information for the given address.
        """
        return self.data[item]


@parser(Specs.ipv4_neigh)
class Ipv4Neigh(IpNeighParser):
    """
    Class to parse ``ip -4 neigh show nud all`` command output.
    """
    pass


@parser(Specs.ipv6_neigh)
class Ipv6Neigh(IpNeighParser):
    """
    Class to parse ``ip -6 neigh show nud all`` command output.
    """
    pass


@parser(Specs.ip_neigh_show)
class IpNeighShow(IpNeighParser):
    """
    Class to parse ``ip neigh show`` or ``ip -s -s neigh show`` command output.
    """
    pass


@parser(Specs.ip_s_link)
class IpLinkInfo(IpAddr):
    """
    This parser parses the output of ``ip -s link`` command, which shows the
    data link layer stats of network devices, like packets received, packets
    dropped, link state, mtu.

    Example output::

        1: enp0s3: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc pfifo_fast state UP mode DEFAULT qlen 1000
                    link/ether 08:00:27:4a:c5:ef brd ff:ff:ff:ff:ff:ff
                    RX: bytes  packets  errors  dropped overrun mcast
                    1113685    2244     0       0       0       0
                    TX: bytes  packets  errors  dropped carrier collsns
                    550754     1407     0       0       0       0


    Resultant output::

        {
            "index": 1,
            "physical_name": null,
            "name": "enp0s3",
            "flags": [
                "BROADCAST",
                "MULTICAST",
                "UP",
                "LOWER_UP"
            ],
            "addr":[]
            "mtu": 1500,
            "qdisc": "pfifo_fast",
            "state": "UP",
            "mode": "DEFAULT",
            "qlen": 1000,
            "mac": "08:00:27:4a:c5:ef",
            "brd": "ff:ff:ff:ff:ff:ff",
            "rx_bytes": 1113685,
            "rx_packets": 2244,
            "rx_errors": 0,
            "rx_dropped": 0,
            "rx_overrun": 0,
            "rx_mcast": 0,
            "tx_bytes": 550754,
            "tx_packets": 1407,
            "tx_errors": 0,
            "tx_dropped": 0,
            "tx_carrier": 0,
            "tx_collsns": 0
        }

    Examples:

        >>> type(ip_link)
        <class 'insights.parsers.ip.IpLinkInfo'>
        >>> for iface in ip_link:
        ...     print 'Interface:', iface['name']
        ...     print 'State:', iface['state']
        ...     print 'RX packets:', iface['rx_packets']
        ...     print 'RX dropped:', iface['rx_dropped']
        ...     print 'TX packets:', iface['tx_packets']
        ...     print 'TX dropped:', iface['tx_dropped']
        ...
        Interface: lo
        State: UNKNOWN
        RX packets: 98
        RX dropped: 0
        TX packets: 10
        TX dropped: 0
        Interface: enp0s3
        State: UP
        RX packets: 2244
        RX dropped: 0
        TX packets: 1407
        TX dropped: 0
        Interface: enp0s8
        State: UP
        RX packets: 1
        RX dropped: 0
        TX packets: 4
        TX dropped: 0
        Interface: enp0s9
        State: UP
        RX packets: 8
        RX dropped: 0
        TX packets: 12
        TX dropped: 0
    """
    pass
